# -*- coding: utf-8 -*-

#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from __future__ import print_function as _
from __future__ import division as _
from __future__ import absolute_import as _
from ..basic_graph_ops import delete_node, disconnect_edge
from .visitors import FindAllUpstreamTerminals

import logging
from coremltools._deps import _HAS_TF_2


def compute_max_rank(graph):
    #  highly inefficient way to calculate the rank of every node
    ret = {}
    # begin at max rank
    for v in graph.keys():
        if len(graph[v].inputs) == 0:
            ret[v] = 0
        else:
            ret[v] = len(graph)

    changes = True
    while changes:
        changes = False
        for v in graph.keys():
            if len(graph[v].inputs) > 0:
                rank = max(ret[i] for i in graph[v].inputs) + 1
                if ret[v] != rank:
                    changes = True
                    ret[v] = rank
    return ret


class CondToWhere(object):
    @staticmethod
    def _search(g, node_name):
        """
        Find the nearest Switch nodes upstream of node_name.
        """
        node = g[node_name]

        switches = (
            FindAllUpstreamTerminals(lambda x: x.op == "Switch")
            .visit(g, node.name)
            .get_result()
        )
        if len(switches) == 0:
            switches = (
                FindAllUpstreamTerminals(
                    lambda x: x.op == "Switch" or x.attr.get("was_switch") is not None
                )
                .visit(g, node.name)
                .get_result()
            )
        return switches

    @staticmethod
    def _fix_found_cond(g, merge, switches):
        """
        Convert a Merge's Switch nodes to Identity ops and the Merge to iff.
        """
        if g[switches[0]].op == "Switch":
            condition_input = g[switches[0]].inputs[1]
        else:
            condition_input = g[switches[0]].attr["was_switch"]

        # convert the merge to a select
        # TensorFlow seems to ensure the condition that the first
        # merge input is the True branch and the second merge input
        # is the false branch.

        # we convert switches to identity, detaching to switch condition
        for s in switches:
            if g[s].op == "Switch":
                g[s].op = "Identity"
                g[s].attr["was_switch"] = g[s].inputs[1]
                # detach input 1: the switch condition
                if g[s].inputs[0] == g[s].inputs[1]:
                    g[s].inputs.pop()
                    g[g[s].inputs[0]].outputs.pop()
                else:
                    disconnect_edge(g, g[s].inputs[1], s)

        # build the final select
        g[merge].op = "iff"
        if not _HAS_TF_2:
            # swap true branch with false branch to get the right semantics for IFF
            g[merge].inputs[0], g[merge].inputs[1] = (
                g[merge].inputs[1],
                g[merge].inputs[0],
            )

        g[merge].inputs = [condition_input] + g[merge].inputs
        g[condition_input].outputs.append(merge)

    def cond_to_where(self, graph):
        stuff_done = False
        g = graph
        ranks = compute_max_rank(graph)
        merges = [a for a in g if g[a].op == "Merge"]
        merges = sorted(merges, key=lambda k: ranks[k])
        if len(merges) == 0:
            return False
        for m in merges:
            logging.debug("Fixing cond at merge location: %s", m)
            switches = self._search(g, m)
            self._fix_found_cond(g, m, switches)
            stuff_done = True

        # delete the extra switches that seem to just lead to identities
        # which then lead nowhere but into control dependencies
        extra_switches = [a for a in g if g[a].op == "Switch"]
        for s in extra_switches:
            if all(
                [g[o].op == "Identity" and len(g[o].outputs) == 0 for o in g[s].outputs]
            ):
                nodes_to_delete = g[s].outputs + [s]
                for d in nodes_to_delete:
                    delete_node(g, d)
                    stuff_done = True
        return stuff_done


def cond_to_where(tfssa):
    for k, v in tfssa.functions.items():
        while True:
            stuff_done = CondToWhere().cond_to_where(v.graph)
            if not stuff_done:
                break



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from coremltools.converters.mil.mil.types.symbolic import any_symbolic
from coremltools.converters.mil.mil import get_new_symbol, get_new_variadic_symbol
from ._op_reqs import *

"""
Random Op Superclass
"""


class RandomDistribution(Operation):
    input_spec = InputSpec(shape=IntTensorInputType(),)

    def __init__(self, **kwargs):
        super(RandomDistribution, self).__init__(**kwargs)

    def type_inference(self):
        if any_symbolic(self.shape.shape):
            # We can't infer any shape if shape has variable length.
            return types.tensor(types.fp32, (get_new_variadic_symbol(),))

        # shape has fixed length here.
        if self.shape.sym_val is None:
            shape = tuple([get_new_symbol() for _ in range(self.shape.shape[0])])
            return types.tensor(types.fp32, shape)

        return types.tensor(types.fp32, tuple(self.shape.sym_val.tolist()))


"""
Random Op Implementation(s)
"""


@register_op(doc_str="")
class random_bernoulli(RandomDistribution):
    r"""
    Returns a tensor with the specified shape, with random values from a Bernoulli
    distribution.
    
    .. math::
       f(k) = \begin{cases}1-p  &\text{if } k = 0\\
                        p    &\text{if } k = 1\end{cases}

    for :math:`k` in :math:`\{0, 1\}`.
    
    Parameters
    ----------
    shape: <K, i32> (Required)
        * Target output tensor shape.
        * ``K`` is the rank of the output tensor.
          ``shape[k] > 0`` for ``k = 0,..., K-1``.
    prob: const<f32> (Optional)
        * The probability of sampling ``1``. Defaults to ``0.5``.
    seed: const<i32> (Optional)
        * Seed to create a reproducible sequence of values across multiple invokes.
    
    Returns
    -------
    <\*, T>
        * A tensor of the given target output shape filled with random values.
    
    See Also
    --------
    random_categorical, random_normal, random_uniform
    """
    
    input_spec = (
        InputSpec(
            shape=IntTensorInputType(),
            prob=FloatInputType(const=True, optional=True),
            seed=IntInputType(const=True, optional=True),
        )
        + RandomDistribution.input_spec
    )

    def default_inputs(self):
        return super().default_inputs() + \
            DefaultInputs(
                seed=-1,
                prob=0.5,
                )

    def __init__(self, **kwargs):
        super(random_bernoulli, self).__init__(**kwargs)


@register_op(doc_str="")
class random_categorical(Operation):
    """
    Returns random values from a categorical distribution.
    
    Parameters
    ----------
    shape: <\*D_in, T>
        * N-dimensional tensor, one of ``logits`` (event log-probabilities) or ``probs``
          (event probabilities). The first ``N - 1`` dimensions specifies distributions,
          and the last dimension represents a vector of probabilities.

    mode: const<str> (Optional)
        One of ``['logits', 'probs']``. Defaults to ``logits``.

    size: const<i32> (Optional)
        Number of samples to draw. Defaults to ``1``.

    seed: const<i32> (Optional)
        Seed to create a reproducible sequence of values across multiple invokes.
    
    Returns
    -------
    <\*D_in[:-1] + [size], T>
        * A tensor of the given target output shape filled with random values.
    
    See Also
    --------
    random_bernoulli, random_normal, random_uniform
    """
    
    input_spec = InputSpec(
        x=TensorInputType(),
        mode=StringInputType(const=True, optional=True),
        size=IntInputType(const=True, optional=True),
        seed=IntInputType(const=True, optional=True),
    )

    def default_inputs(self):
        return DefaultInputs(
            mode="logits",
            size=1,
            seed=-1,
            )

    def __init__(self, **kwargs):
        super(random_categorical, self).__init__(**kwargs)

    def type_inference(self):
        output_shape = self.x.shape[:-1] + (self.size.val,)
        return types.tensor(types.fp32, output_shape)


@register_op(doc_str="")
class random_normal(RandomDistribution):
    r"""
    Returns a tensor with the specified shape, with random values from a normal
    distribution.
    
    Parameters
    ----------
    shape: <K, i32> (Required)
        * Target output tensor shape.
        * ``K`` is the rank of the output tensor.
          ``shape[k] > 0`` for ``k = 0,..., K-1``.
    mean: const<f32> (Optional)
        The mean (center) of the normal distribution. Defaults to 0.0.
    stddev: const<f32> (Optional)
        The standard deviation (width) of the normal distribution. Defaults to ``1.0``.
    seed: const<i32> (Optional)
        Seed to create a reproducible sequence of values across multiple invokes.
    
    Returns
    -------
    <\*, T>
        * A tensor of the given target output shape filled with random values.
    
    See Also
    --------
    random_categorical, random_bernoulli, random_uniform
    """
    
    input_spec = (
        InputSpec(
            shape=IntTensorInputType(),
            mean=FloatInputType(const=True, optional=True),
            stddev=FloatInputType(const=True, optional=True),
            seed=IntInputType(const=True, optional=True),
        )
        + RandomDistribution.input_spec
    )

    def default_inputs(self):
        return super().default_inputs() + \
            DefaultInputs(
                mean=0.,
                stddev=1.,
                seed=-1,
                )

    def __init__(self, **kwargs):
        super(random_normal, self).__init__(**kwargs)


@register_op(doc_str="")
class random_uniform(RandomDistribution):
    r"""
    Returns a tensor with the specified shape with random values from a uniform
    distribution. Samples are uniformly distributed over the half-open interval
    ``[low, high)`` (includes low, but excludes high).
    
    .. math::
       p(x) = \frac{1}{high - low}
    
    For a real number :math:`x`.
    
    When ``high == low``, values of ``low`` will be returned. If ``high < low``,
    the results are officially undefined and may eventually raise an error.
    
    Parameters
    ----------
    shape: <K, i32> (Required)
        * Target output tensor shape.
        * ``K`` is the rank of the output tensor.
          ``shape[k] > 0`` for ``k = 0,..., K-1``.
    low: const<f32> (Optional)
        * Lower boundary of the output interval (inclusive). Defaults to ``0.0``.
    high: const<f32> (Optional)
        * Upper boundary of the output interval (exclusive). Defaults to ``1.0``.
    seed: const<i32> (Optional)
        * Seed to create a reproducible sequence of values across multiple invokes.
    
    Returns
    -------
    <\*, T>
        * A tensor of the given target output shape filled with random values.
    
    See Also
    --------
    random_categorical, random_bernoulli, random_normal
    """
    
    input_spec = (
        InputSpec(
            shape=IntTensorInputType(),
            low=FloatInputType(const=True, optional=True),
            high=FloatInputType(const=True, optional=True),
            seed=IntInputType(const=True, optional=True),
        )
        + RandomDistribution.input_spec
    )

    def default_inputs(self):
        return super().default_inputs() + \
            DefaultInputs(
                low=0.,
                high=1.,
                seed=-1,
                )

    def __init__(self, **kwargs):
        super(random_uniform, self).__init__(**kwargs)



# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import unittest
import numpy as _np
from coremltools._deps import _HAS_SKLEARN

from coremltools.models.utils import evaluate_transformer, _macos_version, _is_macos

if _HAS_SKLEARN:
    from sklearn.preprocessing import StandardScaler
    from coremltools.converters import sklearn as converter


@unittest.skipUnless(
    _is_macos() and _macos_version() >= (10, 13), "Only supported on macOS 10.13+"
)
@unittest.skipIf(not _HAS_SKLEARN, "Missing scikit-learn. Skipping tests.")
class StandardScalerTestCase(unittest.TestCase):
    """
    Unit test class for testing scikit-learn converter.
    """

    def test_random(self):
        # Generate some random data
        X = _np.random.random(size=(50, 3))

        cur_model = StandardScaler()

        output = cur_model.fit_transform(X)

        spec = converter.convert(cur_model, ["a", "b", "c"], "out").get_spec()

        metrics = evaluate_transformer(
            spec,
            [dict(zip(["a", "b", "c"], row)) for row in X],
            [{"out": row} for row in output],
        )

        assert metrics["num_errors"] == 0

    def test_boston(self):
        from sklearn.datasets import load_boston

        scikit_data = load_boston()
        scikit_model = StandardScaler().fit(scikit_data.data)

        spec = converter.convert(
            scikit_model, scikit_data.feature_names, "out"
        ).get_spec()

        input_data = [
            dict(zip(scikit_data.feature_names, row)) for row in scikit_data.data
        ]

        output_data = [{"out": row} for row in scikit_model.transform(scikit_data.data)]

        metrics = evaluate_transformer(spec, input_data, output_data)

        assert metrics["num_errors"] == 0



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from collections import OrderedDict
from itertools import islice

import torch


def _make_ssa_name(name):
    """Converts a symbol name (string) into an SSA name, by prepending '%'.
    Only used for pretty printing the graph.
    """
    return "%" + name


def _ssa_name_list(names):
    """Take a list of symbol names (strings) and return them as SSA names. Only
    used for pretty printing the graph.
    """
    return [_make_ssa_name(x) for x in names]


def _find_new_name(old_name, node_names):
    """Disambiguate a node's name from a list of existing node names by adding
    successively larger integers.
    """
    count = 0
    new_name = old_name + "." + str(count)
    while new_name in node_names:
        count += 1
        new_name = old_name + "." + str(count)
    return new_name


def _replace_in_list(ls, old_val, new_val):
    """Helper function to replace a value in a list."""
    try:
        idx = ls.index(old_val)
    except ValueError:
        pass
    else:
        ls[idx] = new_val


class InternalTorchIRBlock:
    """CoreML internal representation of a torch IR block.

        Arguments:
            raw_block: The torch._C.Block to convert, or None.
            parent: The InternalTorchIRNode this block belongs to.
            nodes: If @raw_block is None, the list of InternalTorchIRNodes in the block
            inputs: If @raw_block is None, the list of input symbols.
            outputs: If @raw_block is None, the list of output symbols.
    """

    def __init__(self, raw_block=None, parent=None, nodes=None, inputs=None, outputs=None):
        self.nodes = []
        node_names = set()
        self.inputs = []
        self.outputs = []
        self.parent = parent

        if raw_block:
            # Add nodes
            for raw_node in raw_block.nodes():
                new_node = InternalTorchIRNode(raw_node, parent=self)
                if new_node.name == new_node.kind:
                    new_node.name = _find_new_name(new_node.name, node_names)
                self.nodes.append(new_node)
                node_names.add(new_node.name)

            # Add inputs
            for inp in raw_block.inputs():
                self.inputs.append(inp.debugName())

            # Add outputs
            for outp in raw_block.outputs():
                self.outputs.append(outp.debugName())
        else:
            self.nodes = nodes
            self.inputs = inputs
            self.outputs = outputs

    def __str__(self, indent=2):
        indent_str = " " * indent
        graph_str = "{}block({}):\n".format(
            indent_str, ", ".join(_ssa_name_list(self.inputs))
        )
        graph_str += "{}\n".format(indent_str).join(
            [x.__str__(indent=indent + 2) for x in self.nodes]
        )
        graph_str += "\n{}return ({})".format(
            indent_str, ", ".join(_ssa_name_list(self.outputs))
        )
        return graph_str

    def __repr__(self):
        return str(self)

    def replace_name(self, old_name, new_name):
        """Replaces all instances of @old_name with @new_name in @self."""

        # Replace graph inputs/outputs
        _replace_in_list(self.inputs, old_name, new_name)
        _replace_in_list(self.outputs, old_name, new_name)

        for node in self.nodes:
            node.replace_name(old_name, new_name)


class InternalTorchIRNode:
    """CoreML internal representation of a torch IR node.
    Can construct itself from a provided torchIR node or manually constructed with
    args for testing.

    See InternalTorchIRGraph for the motivation behind this structure.

        Arguments:
            node: The torch._C.Node to convert, or None.
            parent: The InternalTorchIRGraph/Block this node belongs to.
            attr: If @node is not specified, the dict of named attributes.
            inputs: If @node is not specified, the list of input symbols.
            outputs: If @node is not specified, the list of output symbols.
            kind: If @node is not specified, the kind (op) of the node.
            blocks: If @node is not specified, the list of InternalTorchIRBlock.
    """

    def __init__(
        self, node=None, parent=None, attr=None, inputs=None, outputs=None, kind=None, blocks=None,
    ):
        self.parent = parent
        if node is not None:
            self.inputs = [_input.debugName() for _input in node.inputs()]
            self.outputs = [output.debugName() for output in node.outputs()]
            self.kind = node.kind().split("::")[-1].lower()
            self.blocks = [InternalTorchIRBlock(raw_block=b, parent=self) for b in node.blocks()]
            self.attr = {
                name: getattr(node, node.kindOf(name))(name)
                for name in node.attributeNames()
            }
            if "value" not in self.attr:
                self.attr["value"] = None
            # If the output is boolean, explicitly cast it so type inference
            # will work correctly.
            if len(self.outputs) == 1 and next(node.outputs()).type().str() == "bool":
                self.attr["value"] = bool(self.attr["value"])
        else:
            self.inputs = inputs
            self.outputs = outputs
            self.kind = kind
            self.blocks = blocks if blocks is not None else []
            self.attr = attr if attr is not None else {"value": None}
        # On rare occassions, a node has no outputs. In that case, the node's
        # name will be its kind. However, this no longer guarantees the node's
        # name is unique. It will be up to the graph constructing the node to
        # make sure names are unique.
        self.name = self.outputs[0] if len(self.outputs) > 0 else self.kind

    def __str__(self, indent=2):
        node_str = " " * indent + "{} = {}".format(
            ", ".join(_ssa_name_list(self.outputs)), self.kind
        )
        node_str += "[{}]".format(
            ", ".join(
                ["{}={}".format(n, v) for n, v in self.attr.items() if v is not None]
            )
        )
        node_str += "({})".format(", ".join(_ssa_name_list(self.inputs)))
        for b in self.blocks:
            node_str += "\n" + b.__str__(indent=indent + 2)
        return node_str

    def __repr__(self):
        return str(self)

    def replace_name(self, old_name, new_name):
        """Replaces all instances of @old_name with @new_name in @self."""

        _replace_in_list(self.inputs, old_name, new_name)
        _replace_in_list(self.outputs, old_name, new_name)

        if self.name == old_name:
            self.name = new_name
        for block in self.blocks:
            block.replace_name(old_name, new_name)


class InternalTorchIRGraph:
    """CoreML internal representation of a torch IR graph. A torch._C.Graph
    object is not an ideal structure to use in converting to CoreML. Conversion
    to an InternalTorchIRGraph is inserted between the original graph and the
    final CoreML model to address several issues:
        1. A torch._C.graph is hard to work with. For example, its .inputs()
          and .outputs() functions return iterators, so the only way to
          determine the number of inputs/outputs is by counting to the end.
          There are other examples of why the torch structure is hard to work
          with, and this structure alleviates those isses.
        2. torch._C.graph is an internal API and so we can't count on its
          stability. By inserting a layer in between, we can handle any changes
          to torch._C.graph here and isolate the ops code that processes the
          graph.
        3. torch._C.graph does not expose a Python constructor. This makes
          it impossible to write unit tests that isolate specific ops since
          they have to come from actually converting a PyTorch graph. With an
          internal structure, we can directly build the test cases we need for
          unit testing.

        Arguments:
            raw_graph: raw_graph: The torch._C.Graph to convert, or None.
            params_dict: A dictionary mapping graph parameter names to tensors.
                Must be given if @raw_graph is not None.
            input_values: A list of inputs to the graph. Must be given is
                @raw_graph if not None.
            cut_at_symbols: The list of desired outputs from the graph. Symbols
                must be present in the graph. For debugging use only. Can only
                be given if @raw_graph is not None.
            nodes: If @raw_graph is None, the list of InternalTorchIRNodes in
                the graph.
            params: If @raw_graph is None, the dict mapping parameter names to
                their numpy value.
            inputs: If @raw_graph is None, the OrderedDict mapping input names
                to their example values.
            outputs: If @raw_graph is None, the list of outputs from the graph.
    """

    def __init__(
        self, raw_graph=None, params_dict=None, input_values=None, cut_at_symbols=None, nodes=None, params=None, inputs=None, outputs=None,
    ):
        self.nodes = []
        node_names = set()
        self.params = {}
        self.inputs = OrderedDict()
        self.outputs = []

        if raw_graph is not None:
            # Add nodes
            for raw_node in raw_graph.nodes():
                new_node = InternalTorchIRNode(raw_node, parent=self)
                if new_node.name == new_node.kind:
                    new_node.name = _find_new_name(new_node.name, node_names)
                self.nodes.append(new_node)
                node_names.add(new_node.name)

            # Add params
            for name, param in params_dict.items():
                value = param.detach().cpu().numpy()
                self.params[name] = value

            # Add inputs
            for index, _input in enumerate(islice(raw_graph.inputs(), len(input_values))):
                name = _input.debugName()
                value = input_values[index]
                self.inputs[name] = value

            # Add outputs, cutting if @cut_at_symbols is set
            output_names = cut_at_symbols
            if output_names is None:
                output_names = [x.debugName() for x in raw_graph.outputs()]
            for output in output_names:
                self.outputs.append(output)
        else:
            self.nodes = nodes
            self.params = params
            self.inputs = inputs
            self.outputs = outputs

    def __str__(self):
        graph_str = "graph(\n"
        graph_str += self._format_inputs(self.inputs, unpack=True)
        graph_str += self._format_inputs(self.params)
        graph_str += "):\n"
        graph_str += "\n".join([str(x) for x in self.nodes]) + "\n"
        graph_str += "return ({})".format(", ".join(_ssa_name_list(self.outputs)))
        return graph_str

    def _format_inputs(self, inputs, unpack=False):
        def tensor_str(x):
            return "Tensor{}".format(
                tuple(list(x.shape.shape if unpack else x.shape) + [str(x.dtype)])
            )

        inp_str = ""
        for k, v in inputs.items():
            if isinstance(v, (tuple, list)):
                shape_str = "({})".format(", ".join([tensor_str(x) for x in v]))
            else:
                shape_str = tensor_str(v)
            inp_str += "    {} : {},\n".format(_make_ssa_name(k), shape_str)
        return inp_str

    def __repr__(self):
        return str(self)

    def replace_name(self, old_name, new_name):
        """Replaces all instances of @old_name with @new_name in @self."""

        # Replace graph inputs/outputs
        _replace_in_list(self.inputs, old_name, new_name)
        _replace_in_list(self.outputs, old_name, new_name)

        for node in self.nodes:
            node.replace_name(old_name, new_name)



# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from . import _SVR as _SVR

from ..._deps import _HAS_SKLEARN
from ...models import MLModel as _MLModel

if _HAS_SKLEARN:
    from ._sklearn_util import check_fitted
    from sklearn.svm import NuSVR as _NuSVR
    from . import _sklearn_util

    sklearn_class = _NuSVR

model_type = "regressor"


def convert(model, feature_names, target):
    """Convert a Nu Support Vector Regression (NuSVR) model to the protobuf spec.
    Parameters
    ----------
    model: NuSVR
        A trained NuSVR encoder model.

    feature_names: [str]
        Name of the input columns.

    target: str
        Name of the output column.

    Returns
    -------
    model_spec: An object of type Model_pb.
        Protobuf representation of the model
    """
    if not (_HAS_SKLEARN):
        raise RuntimeError(
            "scikit-learn not found. scikit-learn conversion API is disabled."
        )

    _sklearn_util.check_expected_type(model, _NuSVR)
    return _SVR.convert(model, feature_names, target)


def get_input_dimension(model):
    if not (_HAS_SKLEARN):
        raise RuntimeError(
            "scikit-learn not found. scikit-learn conversion API is disabled."
        )

    check_fitted(model, lambda m: hasattr(m, "support_vectors_"))
    return _SVR.get_input_dimension(model)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: NamedParameters.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='NamedParameters.proto',
  package='CoreML.Specification',
  syntax='proto3',
  serialized_pb=_b('\n\x15NamedParameters.proto\x12\x14\x43oreML.Specification\"0\n\nInt32Range\x12\x10\n\x08minValue\x18\x01 \x01(\x05\x12\x10\n\x08maxValue\x18\x02 \x01(\x05\"\x1a\n\x08Int32Set\x12\x0e\n\x06values\x18\x01 \x03(\x05\"0\n\nFloatRange\x12\x10\n\x08minValue\x18\x01 \x01(\x02\x12\x10\n\x08maxValue\x18\x02 \x01(\x02\"\x99\x01\n\x0eInt32Parameter\x12\x14\n\x0c\x64\x65\x66\x61ultValue\x18\x01 \x01(\x05\x12\x31\n\x05range\x18\n \x01(\x0b\x32 .CoreML.Specification.Int32RangeH\x00\x12-\n\x03set\x18\x0b \x01(\x0b\x32\x1e.CoreML.Specification.Int32SetH\x00\x42\x0f\n\rAllowedValues\"j\n\x0e\x46loatParameter\x12\x14\n\x0c\x64\x65\x66\x61ultValue\x18\x01 \x01(\x02\x12\x31\n\x05range\x18\n \x01(\x0b\x32 .CoreML.Specification.FloatRangeH\x00\x42\x0f\n\rAllowedValues\"\x93\x01\n\tParameter\x12>\n\x0eint32Parameter\x18\x01 \x01(\x0b\x32$.CoreML.Specification.Int32ParameterH\x00\x12>\n\x0e\x66loatParameter\x18\x02 \x01(\x0b\x32$.CoreML.Specification.FloatParameterH\x00\x42\x06\n\x04Type\"l\n\x0eNamedParameter\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x18\n\x10shortDescription\x18\x02 \x01(\t\x12\x32\n\tparameter\x18\x03 \x01(\x0b\x32\x1f.CoreML.Specification.ParameterB\x02H\x03\x62\x06proto3')
)




_INT32RANGE = _descriptor.Descriptor(
  name='Int32Range',
  full_name='CoreML.Specification.Int32Range',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='minValue', full_name='CoreML.Specification.Int32Range.minValue', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='maxValue', full_name='CoreML.Specification.Int32Range.maxValue', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=47,
  serialized_end=95,
)


_INT32SET = _descriptor.Descriptor(
  name='Int32Set',
  full_name='CoreML.Specification.Int32Set',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='values', full_name='CoreML.Specification.Int32Set.values', index=0,
      number=1, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=97,
  serialized_end=123,
)


_FLOATRANGE = _descriptor.Descriptor(
  name='FloatRange',
  full_name='CoreML.Specification.FloatRange',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='minValue', full_name='CoreML.Specification.FloatRange.minValue', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='maxValue', full_name='CoreML.Specification.FloatRange.maxValue', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=125,
  serialized_end=173,
)


_INT32PARAMETER = _descriptor.Descriptor(
  name='Int32Parameter',
  full_name='CoreML.Specification.Int32Parameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='defaultValue', full_name='CoreML.Specification.Int32Parameter.defaultValue', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='range', full_name='CoreML.Specification.Int32Parameter.range', index=1,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='set', full_name='CoreML.Specification.Int32Parameter.set', index=2,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='AllowedValues', full_name='CoreML.Specification.Int32Parameter.AllowedValues',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=176,
  serialized_end=329,
)


_FLOATPARAMETER = _descriptor.Descriptor(
  name='FloatParameter',
  full_name='CoreML.Specification.FloatParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='defaultValue', full_name='CoreML.Specification.FloatParameter.defaultValue', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='range', full_name='CoreML.Specification.FloatParameter.range', index=1,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='AllowedValues', full_name='CoreML.Specification.FloatParameter.AllowedValues',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=331,
  serialized_end=437,
)


_PARAMETER = _descriptor.Descriptor(
  name='Parameter',
  full_name='CoreML.Specification.Parameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='int32Parameter', full_name='CoreML.Specification.Parameter.int32Parameter', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='floatParameter', full_name='CoreML.Specification.Parameter.floatParameter', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='Type', full_name='CoreML.Specification.Parameter.Type',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=440,
  serialized_end=587,
)


_NAMEDPARAMETER = _descriptor.Descriptor(
  name='NamedParameter',
  full_name='CoreML.Specification.NamedParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='CoreML.Specification.NamedParameter.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='shortDescription', full_name='CoreML.Specification.NamedParameter.shortDescription', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='parameter', full_name='CoreML.Specification.NamedParameter.parameter', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=589,
  serialized_end=697,
)

_INT32PARAMETER.fields_by_name['range'].message_type = _INT32RANGE
_INT32PARAMETER.fields_by_name['set'].message_type = _INT32SET
_INT32PARAMETER.oneofs_by_name['AllowedValues'].fields.append(
  _INT32PARAMETER.fields_by_name['range'])
_INT32PARAMETER.fields_by_name['range'].containing_oneof = _INT32PARAMETER.oneofs_by_name['AllowedValues']
_INT32PARAMETER.oneofs_by_name['AllowedValues'].fields.append(
  _INT32PARAMETER.fields_by_name['set'])
_INT32PARAMETER.fields_by_name['set'].containing_oneof = _INT32PARAMETER.oneofs_by_name['AllowedValues']
_FLOATPARAMETER.fields_by_name['range'].message_type = _FLOATRANGE
_FLOATPARAMETER.oneofs_by_name['AllowedValues'].fields.append(
  _FLOATPARAMETER.fields_by_name['range'])
_FLOATPARAMETER.fields_by_name['range'].containing_oneof = _FLOATPARAMETER.oneofs_by_name['AllowedValues']
_PARAMETER.fields_by_name['int32Parameter'].message_type = _INT32PARAMETER
_PARAMETER.fields_by_name['floatParameter'].message_type = _FLOATPARAMETER
_PARAMETER.oneofs_by_name['Type'].fields.append(
  _PARAMETER.fields_by_name['int32Parameter'])
_PARAMETER.fields_by_name['int32Parameter'].containing_oneof = _PARAMETER.oneofs_by_name['Type']
_PARAMETER.oneofs_by_name['Type'].fields.append(
  _PARAMETER.fields_by_name['floatParameter'])
_PARAMETER.fields_by_name['floatParameter'].containing_oneof = _PARAMETER.oneofs_by_name['Type']
_NAMEDPARAMETER.fields_by_name['parameter'].message_type = _PARAMETER
DESCRIPTOR.message_types_by_name['Int32Range'] = _INT32RANGE
DESCRIPTOR.message_types_by_name['Int32Set'] = _INT32SET
DESCRIPTOR.message_types_by_name['FloatRange'] = _FLOATRANGE
DESCRIPTOR.message_types_by_name['Int32Parameter'] = _INT32PARAMETER
DESCRIPTOR.message_types_by_name['FloatParameter'] = _FLOATPARAMETER
DESCRIPTOR.message_types_by_name['Parameter'] = _PARAMETER
DESCRIPTOR.message_types_by_name['NamedParameter'] = _NAMEDPARAMETER
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Int32Range = _reflection.GeneratedProtocolMessageType('Int32Range', (_message.Message,), dict(
  DESCRIPTOR = _INT32RANGE,
  __module__ = 'NamedParameters_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.Int32Range)
  ))
_sym_db.RegisterMessage(Int32Range)

Int32Set = _reflection.GeneratedProtocolMessageType('Int32Set', (_message.Message,), dict(
  DESCRIPTOR = _INT32SET,
  __module__ = 'NamedParameters_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.Int32Set)
  ))
_sym_db.RegisterMessage(Int32Set)

FloatRange = _reflection.GeneratedProtocolMessageType('FloatRange', (_message.Message,), dict(
  DESCRIPTOR = _FLOATRANGE,
  __module__ = 'NamedParameters_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.FloatRange)
  ))
_sym_db.RegisterMessage(FloatRange)

Int32Parameter = _reflection.GeneratedProtocolMessageType('Int32Parameter', (_message.Message,), dict(
  DESCRIPTOR = _INT32PARAMETER,
  __module__ = 'NamedParameters_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.Int32Parameter)
  ))
_sym_db.RegisterMessage(Int32Parameter)

FloatParameter = _reflection.GeneratedProtocolMessageType('FloatParameter', (_message.Message,), dict(
  DESCRIPTOR = _FLOATPARAMETER,
  __module__ = 'NamedParameters_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.FloatParameter)
  ))
_sym_db.RegisterMessage(FloatParameter)

Parameter = _reflection.GeneratedProtocolMessageType('Parameter', (_message.Message,), dict(
  DESCRIPTOR = _PARAMETER,
  __module__ = 'NamedParameters_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.Parameter)
  ))
_sym_db.RegisterMessage(Parameter)

NamedParameter = _reflection.GeneratedProtocolMessageType('NamedParameter', (_message.Message,), dict(
  DESCRIPTOR = _NAMEDPARAMETER,
  __module__ = 'NamedParameters_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.NamedParameter)
  ))
_sym_db.RegisterMessage(NamedParameter)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('H\003'))
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-

#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from __future__ import print_function as _
from __future__ import division as _
from __future__ import absolute_import as _

from coremltools.converters.mil.mil.passes.pass_registry import register_pass
from coremltools.converters.mil.mil import Builder as mb
import numpy as np


def _check_child_op_type(op, child_op_type):
    """
    :param op: operation
    :param child_op_type: str
    :return: Return True if op has 1 child and type of that child matches child_op_type
    """
    if len(op.outputs) != 1:
        return False
    child_ops = list(op.outputs[0].child_ops)
    if len(child_ops) != 1:
        return False
    if child_ops[0].op_type == child_op_type:
        return True
    return False


def _check_var_scalar_value(x, val, tol=1e-3):
    """
    :param x: var
    :param val: a scalar value
    :return: True if the value of var is equal to val otherwise return False
    """
    if x.val is None:
        return False
    if not isinstance(x.val, (np.ndarray, np.generic)):
        return False

    if isinstance(x.val, np.ndarray):
        if x.val.size != 1:
            return False
        x_val = x.val[:][0]
    else:
        x_val = x.val

    if abs(x_val - val) < tol:
        return True
    return False


def try_to_transform(onehot_op, block):
    root_var = onehot_op.indices

    # check that the output of the onehot op is not a block output
    if onehot_op.outputs[0] in block.outputs:
        return False

    # check that onehot op has axis=-1, on_value=1 and off_value=0
    # and constant one_hot_vector_size
    axis = onehot_op.axis.val
    if axis is None:
        return False
    if onehot_op.indices.shape is None:
        return False
    rank = len(onehot_op.indices.shape)
    if axis >= 0:
        axis -= rank
    if axis != -1:
        return False
    if not _check_var_scalar_value(onehot_op.on_value, 1):
        return False
    if not _check_var_scalar_value(onehot_op.off_value, 0):
        return False
    if onehot_op.one_hot_vector_size.val is None:
        return False

    # checks for the following matmul op
    if not _check_child_op_type(onehot_op, "matmul"):
        return False
    matmul_op = list(onehot_op.outputs[0].child_ops)[0]
    if matmul_op.x != onehot_op.outputs[0]:
        return False
    if matmul_op.transpose_x.val or matmul_op.transpose_y.val:
        return False
    W_var = matmul_op.y
    if W_var.val is None:
        return False
    if len(W_var.val.shape) != 2:
        return False

    # remove onehot and matmul and replace with gather op
    out_name = matmul_op.outputs[0].name
    x = mb.gather(x=W_var, indices=root_var, axis=0, name=out_name, before_op=matmul_op)

    matmul_op.enclosing_block.replace_uses_of_var_after_op(
        anchor_op=matmul_op, old_var=matmul_op.outputs[0], new_var=x
    )
    # Remove all the ops at once
    block.remove_ops([onehot_op, matmul_op])
    return True


def fuse_onehot_matmul_to_gather_block(block):
    fusion_status = False
    for i, op in enumerate(list(block.operations)):
        for b in op.blocks:
            block_changed = True
            while block_changed:
                block_changed = fuse_onehot_matmul_to_gather_block(b)
        if len(op.blocks) > 0:
            # This op can't be pow
            continue

        # start pattern match if one_hot op is encountered
        if op.op_type == "one_hot":
            with block:
                fusion_status = try_to_transform(op, block)
            # has to break as the downstream iterator is affected.
            if fusion_status:
                return fusion_status
    return fusion_status


@register_pass(namespace="common")
def fuse_onehot_matmul_to_gather(prog):
    """
    Detect if onehot (axis=-1, on_value=1, off_value=0) is followed by a matmul op (no bias),
    then they can be replaced by a gather op.

    Input:
        %2 = one_hot(%1, on_value=1, off_value=0, axis=-1)
        %3 = const() # rank 2
        %4  = matmul(%2, %3)

    Output:
        %4 = gather(%3, %2, axis=0)

    """
    for f_name, f in prog.functions.items():
        block_changed = True
        while block_changed:
            block_changed = fuse_onehot_matmul_to_gather_block(f)



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import scipy
from scipy import special
from coremltools.converters.mil import testing_reqs
from coremltools.converters.mil.testing_reqs import *

from .testing_utils import run_compare_builder

backends = testing_reqs.backends


class TestClampedReLU:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        t = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        input_placeholders = {
            "x": mb.placeholder(shape=t.shape),
        }
        input_values = {"x": t}

        def build(x):
            return mb.clamped_relu(x=x, alpha=2.0, beta=1.0)

        expected_output_types = (2, 3, types.fp32)
        expected_outputs = np.array([[-2, 1, -6], [1, -10, 1]], dtype=np.float32)

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @ssa_fn
    def test_builder_eval(self):
        x_val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        v = mb.clamped_relu(x=x_val, alpha=2.0, beta=1.0)

        x = np.minimum(np.maximum(x_val, 0), 1.0)
        y = np.minimum(np.minimum(x_val, 0) * 2.0, 1.0)
        assert is_close(x + y, v.val)

    @pytest.mark.parametrize(
        "use_cpu_only, backend, dim, alpha, beta",
        itertools.product([True], backends, [2, 4, 8], [2.0, 3.0], [4.0, 5.0]),
    )
    def test_builder_to_backend_stress(self, use_cpu_only, backend, dim, alpha, beta):
        shape_x = np.array([dim, dim])
        x_val = np.random.rand(*shape_x)
        input_placeholders = {"x": mb.placeholder(shape=x_val.shape)}
        input_values = {"x": x_val}

        def build(x):
            return [mb.clamped_relu(x=x, alpha=alpha, beta=beta)]

        x = np.minimum(np.maximum(x_val, 0), 1.0)
        y = np.minimum(np.minimum(x_val, 0) * 2.0, 1.0)

        expected_outputs = [x + y]
        expected_output_types = [o.shape[:] + (types.fp32,) for o in expected_outputs]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs=expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
        )


class TestELU:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        t = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        input_placeholders = {
            "x": mb.placeholder(shape=t.shape),
        }
        input_values = {"x": t}

        def build(x):
            return mb.elu(x=x, alpha=2.0)

        expected_output_types = (2, 3, types.fp32)
        expected_outputs = np.array(
            [[-1.2642411, 2.0, -1.9004259], [4.0, -1.9865241, 6.0]], dtype=np.float32
        )

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @ssa_fn
    def test_builder_eval(self):
        x_val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        v = mb.elu(x=x_val, alpha=2.0)

        b = np.copy(x_val)
        b[b < 0] = 2.0 * (np.exp(b[b < 0]) - 1)

        assert is_close(b, v.val)


class TestGeLU:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        t = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        input_placeholders = {
            "x": mb.placeholder(shape=t.shape),
        }
        input_values = {"x": t}

        def build(x):
            return mb.gelu(x=x)

        expected_output_types = (2, 3, types.fp32)
        expected_outputs = np.array(
            [
                [-1.58691406e-01, 1.95410156e00, -4.04968858e-03],
                [3.99987316e00, -1.49011612e-06, 6.00000000e00],
            ],
            dtype=np.float32,
        )

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
            atol=1e-3,
            rtol=1e-3,
        )

    @ssa_fn
    def test_builder_eval(self):
        x_val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)

        mode = "TANH_APPROXIMATION"
        v = mb.gelu(x=x_val, mode=mode)
        a = np.sqrt(2 / np.pi) * (x_val + 0.044715 * np.power(x_val, 3))
        out = 0.5 * x_val * (1 + np.tanh(a))
        assert is_close(out, v.val)

        mode = "SIGMOID_APPROXIMATION"
        v = mb.gelu(x=x_val, mode=mode)
        out = x_val * (1 / (1 + np.exp(-(1.702 * x_val))))
        assert is_close(out, v.val)

        v = mb.gelu(x=x_val)
        out = 0.5 * x_val * (1 + scipy.special.erf(x_val / np.sqrt(2)))
        assert is_close(out, v.val)

    @pytest.mark.parametrize(
        "use_cpu_only, backend, dim, mode",
        itertools.product(
            [True, False],
            backends,
            [2, 6],
            ["EXACT", "TANH_APPROXIMATION", "SIGMOID_APPROXIMATION"],
        ),
    )
    def test_builder_to_backend_stress(self, use_cpu_only, backend, dim, mode):
        shape = np.array([dim, dim])
        x_val = np.random.rand(*shape)
        input_placeholders = {"x": mb.placeholder(shape=x_val.shape)}
        input_values = {"x": x_val}

        def build(x):
            return [mb.gelu(x=x, mode=mode)]

        if mode == "TANH_APPROXIMATION":
            a = np.sqrt(2 / np.pi) * (x_val + 0.044715 * np.power(x_val, 3))
            out = 0.5 * x_val * (1 + np.tanh(a))
        elif mode == "SIGMOID_APPROXIMATION":
            out = x_val * (1 / (1 + np.exp(-(1.702 * x_val))))
        else:
            out = 0.5 * x_val * (1 + scipy.special.erf(x_val / np.sqrt(2)))

        expected_outputs = [out]
        expected_output_types = [o.shape[:] + (types.fp32,) for o in expected_outputs]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs=expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
            atol=1e-3,
            rtol=1e-3,
        )


class TestLeakyReLU:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        t = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        input_placeholders = {
            "x": mb.placeholder(shape=t.shape),
        }
        input_values = {"x": t}

        def build(x):
            return mb.leaky_relu(x=x, alpha=2.0)

        expected_output_types = (2, 3, types.fp32)
        expected_outputs = np.array([[-2, 2, -6], [4, -10, 6]], dtype=np.float32)

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @ssa_fn
    def test_builder_eval(self):
        x_val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        v = mb.leaky_relu(x=x_val, alpha=2.0)

        b = np.copy(x_val)
        b[b < 0] *= 2.0
        assert is_close(b, v.val)


class TestLinearActivation:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        t = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        input_placeholders = {"x": mb.placeholder(shape=t.shape)}
        input_values = {"x": t}

        def build(x):
            return mb.linear_activation(x=x, alpha=2.0, beta=3.0)

        expected_output_types = (2, 3, types.fp32)
        expected_outputs = np.array([[1, 7, -3], [11, -7, 15]], dtype=np.float32)

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @ssa_fn
    def test_builder_eval(self):
        x_val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        v = mb.linear_activation(x=x_val, alpha=2.0, beta=3.0)
        assert is_close(x_val * 2.0 + 3.0, v.val)

    @pytest.mark.parametrize(
        "use_cpu_only, backend, dim",
        itertools.product([True, False], backends, [2, 4, 8]),
    )
    def test_builder_to_backend_stress(self, use_cpu_only, backend, dim):
        shape = np.array([dim, dim])
        x_val = np.random.rand(*shape)
        alpha = np.random.uniform()
        beta = np.random.uniform()
        input_placeholders = {
            "x": mb.placeholder(shape=x_val.shape),
        }
        input_values = {"x": x_val}

        def build(x):
            return [mb.linear_activation(x=x, alpha=alpha, beta=beta)]

        expected_outputs = [x_val * alpha + beta]
        expected_output_types = [o.shape[:] + (types.fp32,) for o in expected_outputs]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs=expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
        )


# TODO (rdar://59954690): Broken when there is 1 channel
class TestPReLU:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        t = np.array([[[[-1, 3, 6]], [[-1, 2, -3]], [[4, -5, 6]]]], dtype=np.float32)
        input_placeholders = {"x": mb.placeholder(shape=t.shape)}
        input_values = {"x": t}

        def build(x):
            return mb.prelu(x=x, alpha=np.array([1, 2, 3], dtype=np.float32))

        expected_output_types = (1, 3, 1, 3, types.fp32)
        expected_outputs = np.array(
            [[[[-1, 3, 6]], [[-2, 2, -6]], [[4, -15, 6]]]], dtype=np.float32
        )

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @ssa_fn
    def test_builder_eval(self):
        x_val = np.array([[[[-1, 3, 6]], [[-1, 2, -3]], [[4, -5, 6]]]], dtype=np.float32)
        alpha = np.array([1, 2, 3], dtype=np.float32)
        v = mb.prelu(x=x_val, alpha=alpha)

        alpha_br = alpha

        for i in range(1, len(x_val.shape)):
            alpha_br = np.expand_dims(alpha_br, i)

        x_pos = np.maximum(x_val, 0)
        b = np.minimum(x_val, 0)

        assert is_close(x_pos + b * alpha_br, v.val)

    @ssa_fn
    def test_builder_eval1(self):
        x_val = np.array([[[-1, 3, 6]], [[-1, 2, -3]], [[4, -5, 6]]], dtype=np.float32)
        with pytest.raises(ValueError, match=r".* dimension 1 .*"):
            v = mb.prelu(x=x_val, alpha=np.array([1, 2], dtype=np.float32))

    @ssa_fn
    def test_builder_eval2(self):
        x_val = np.array([[[-1, 3, 6]], [[-1, 2, -3]], [[4, -5, 6]]], dtype=np.float32)
        with pytest.raises(ValueError, match=r"alpha .* rank 1"):
            v = mb.prelu(x=x_val, alpha=np.array([[1, 2, 3]], dtype=np.float32))

    @ssa_fn
    def test_builder_eval3(self):
        x_val = np.array([[[-1, 3, 6]], [[-1, 2, -3]], [[4, -5, 6]]], dtype=np.float32)
        with pytest.raises(ValueError, match=r"x .* rank 3"):
            v = mb.prelu(x=[1], alpha=np.array([[1, 2, 3]], dtype=np.float32))

    @pytest.mark.parametrize(
        "use_cpu_only, backend, dim, chan",
        itertools.product([True, False], backends, [1, 2, 4, 8], [2, 3, 4]),
    )
    def test_builder_to_backend_stress(self, use_cpu_only, backend, dim, chan):
        shape = np.array([1, chan, dim, dim])
        x_val = np.random.rand(*shape)
        alpha_val = np.random.rand(chan).astype(np.float32)

        input_placeholders = {"x": mb.placeholder(shape=x_val.shape)}
        input_values = {"x": x_val}

        def build(x):
            return [mb.prelu(x=x, alpha=alpha_val)]

        alpha_br = np.copy(alpha_val)
        for i in range(1, len(x_val.shape)-1):
            alpha_br = np.expand_dims(alpha_br, i)
        x_pos = np.maximum(x_val, 0)
        b = np.minimum(x_val, 0)

        expected_outputs = [x_pos + b * alpha_br]
        expected_output_types = [o.shape[:] + (types.fp32,) for o in expected_outputs]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs=expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
        )


class TestReLU:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        t = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        input_placeholders = {"x": mb.placeholder(shape=t.shape)}
        input_values = {"x": t}

        def build(x):
            return mb.relu(x=x)

        expected_output_types = (2, 3, types.fp32)
        expected_outputs = np.array([[0, 2, 0], [4, 0, 6]], dtype=np.float32)

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @ssa_fn
    def test_builder_eval(self):
        x_val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        v = mb.relu(x=x_val)
        assert is_close(np.maximum(x_val, 0), v.val)


class TestReLU6:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        t = np.array([[-1, 7, -3], [4, -5, 8]], dtype=np.float32)
        input_placeholders = {"x": mb.placeholder(shape=t.shape)}
        input_values = {"x": t}

        def build(x):
            return mb.relu6(x=x)

        expected_output_types = (2, 3, types.fp32)
        expected_outputs = np.array([[0, 6, 0], [4, 0, 6]], dtype=np.float32)

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @ssa_fn
    def test_builder_eval(self):
        x_val = np.array([[-1, 7, -3], [4, -5, 8]], dtype=np.float32)
        v = mb.relu6(x=x_val)
        assert is_close(np.minimum(np.maximum(x_val, 0), 6), v.val)


class TestScaledTanh:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        t = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        input_placeholders = {"x": mb.placeholder(shape=t.shape)}
        input_values = {"x": t}

        def build(x):
            return mb.scaled_tanh(x=x, alpha=2.0, beta=1.0)

        expected_output_types = (2, 3, types.fp32)
        expected_outputs = np.array(
            [[-1.5231884, 1.9280552, -1.9901096], [1.9986587, -1.9998184, 1.9999754]],
            dtype=np.float32,
        )

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @ssa_fn
    def test_builder_eval(self):
        x_val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        v = mb.scaled_tanh(x=x_val, alpha=2.0, beta=1.0)
        assert is_close(2.0 * np.tanh(x_val * 1.0), v.val)

    @pytest.mark.parametrize(
        "use_cpu_only, backend, dim, alpha, beta",
        itertools.product([True], backends, [2, 4, 8], [2.0, 3.0], [4.0, 5.0]),
    )
    def test_builder_to_backend_stress(self, use_cpu_only, backend, dim, alpha, beta):
        shape_x = np.array([dim, dim])
        x_val = np.random.rand(*shape_x)
        input_placeholders = {"x": mb.placeholder(shape=x_val.shape)}
        input_values = {"x": x_val}

        def build(x):
            return [mb.scaled_tanh(x=x, alpha=alpha, beta=beta)]

        expected_outputs = [alpha * np.tanh(x_val * beta)]
        expected_output_types = [o.shape[:] + (types.fp32,) for o in expected_outputs]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs=expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
        )


class TestSigmoid:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        t = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        input_placeholders = {"x": mb.placeholder(shape=t.shape)}
        input_values = {"x": t}

        def build(x):
            return mb.sigmoid(x=x)

        expected_output_types = (2, 3, types.fp32)
        expected_outputs = np.array(
            [
                [0.2689414213699951, 0.8807970779778823, 0.04742587],
                [0.98201376, 0.00669285, 0.9975274],
            ],
            dtype=np.float32,
        )

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @ssa_fn
    def test_builder_eval(self):
        x_val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        v = mb.sigmoid(x=x_val)
        assert is_close(1 / (1 + np.exp(-x_val)), v.val)


class TestSigmoidHard:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        t = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        input_placeholders = {"x": mb.placeholder(shape=t.shape)}
        input_values = {"x": t}

        def build(x):
            return mb.sigmoid_hard(x=x, alpha=1.0, beta=2.0)

        expected_output_types = (2, 3, types.fp32)
        expected_outputs = np.array(
            [[1.0, 1.0, 0.0], [1.0, 0.0, 1.0]], dtype=np.float32
        )

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @ssa_fn
    def test_builder_eval(self):
        x_val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        alpha = 1.0
        beta = 2.0
        v = mb.sigmoid_hard(x=x_val, alpha=alpha, beta=beta)
        assert is_close(np.minimum(np.maximum((alpha * x_val) + beta, 0), 1), v.val)

    @pytest.mark.parametrize(
        "use_cpu_only, backend, dim, alpha, beta",
        itertools.product([True], backends, [2, 4, 8], [2.0, 3.0], [4.0, 5.0]),
    )
    def test_builder_to_backend_stress(self, use_cpu_only, backend, dim, alpha, beta):
        shape_x = np.array([dim, dim])
        x_val = np.random.rand(*shape_x)
        input_placeholders = {"x": mb.placeholder(shape=x_val.shape)}
        input_values = {"x": x_val}

        def build(x):
            return [mb.sigmoid_hard(x=x, alpha=alpha, beta=beta)]

        expected_outputs = [np.minimum(np.maximum((alpha * x_val) + beta, 0), 1)]
        expected_output_types = [o.shape[:] + (types.fp32,) for o in expected_outputs]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs=expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
        )


class TestSoftplus:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        t = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        input_placeholders = {"x": mb.placeholder(shape=t.shape)}
        input_values = {"x": t}

        def build(x):
            return mb.softplus(x=x)

        expected_output_types = (2, 3, types.fp32)
        expected_outputs = np.array(
            [[0.31326166, 2.126928, 0.04858733], [4.01815, 0.00671535, 6.0024757]],
            dtype=np.float32,
        )

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @ssa_fn
    def test_builder_eval(self):
        x_val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        v = mb.softplus(x=x_val)
        assert is_close(
            np.log(1 + np.exp(-np.abs(x_val))) + np.maximum(x_val, 0), v.val
        )


# TODO (rdar://59954690): NN Segfaults when converting from MIL ParametricSoftplus layer
# No torch test because there is no direct torch translation to this layer
class TestSoftplusParametric:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        t = np.array([[[[-1, 3, 6]], [[-1, 2, -3]], [[4, -5, 6]]]], dtype=np.float32)
        input_placeholders = {"x": mb.placeholder(shape=t.shape)}
        input_values = {"x": t}

        def build(x):
            return mb.softplus_parametric(
                x=x,
                alpha=np.array([1, 2, 3], dtype=np.float32),
                beta=np.array([4, 5, 6], dtype=np.float32),
            )

        expected_output_types = (1, 3, 1, 3, types.fp32)
        expected_outputs = np.array(
            [[
                [[1.8142700e-02, 1.2000000e01, 2.4000000e01]],
                [[1.3427734e-02, 2.0000000e01, 7.1525574e-07]],
                [[7.2000000e01, 0.0000000e00, 1.0800000e02]],
            ]],
            dtype=np.float32,
        )

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @ssa_fn
    def test_builder_eval(self):
        x_val = np.array([[[[-1, 3, 6]], [[-1, 2, -3]], [[4, -5, 6]]]], dtype=np.float32)
        v = mb.softplus_parametric(
            x=x_val,
            alpha=np.array([1, 2, 3], dtype=np.float32),
            beta=np.array([4, 5, 6], dtype=np.float32),
        )

        alpha_br = np.array([1, 2, 3], dtype=np.float32)
        beta_br = np.array([4, 5, 6], dtype=np.float32)
        for i in range(1, len(x_val.shape)):
            alpha_br = np.expand_dims(alpha_br, i)
            beta_br = np.expand_dims(beta_br, i)
        out = alpha_br * np.log(np.exp(x_val * beta_br) + 1)

        assert is_close(out, v.val)

    @ssa_fn
    def test_builder_eval2(self):
        x_val = np.array([[[-1, 3, 6]], [[-1, 2, -3]], [[4, -5, 6]]], dtype=np.float32)
        with pytest.raises(ValueError, match=r".* dimension 1 .*"):
            v = mb.softplus_parametric(
                x=x_val,
                alpha=np.array([1, 2], dtype=np.float32),
                beta=np.array([4, 5, 6], dtype=np.float32),
            )

    @ssa_fn
    def test_builder_eval3(self):
        x_val = np.array([[[-1, 3, 6]], [[-1, 2, -3]], [[4, -5, 6]]], dtype=np.float32)
        with pytest.raises(ValueError, match=r"alpha .* rank 1"):
            v = mb.softplus_parametric(
                x=x_val,
                alpha=np.array([[1, 2, 3]], dtype=np.float32),
                beta=np.array([4, 5, 6], dtype=np.float32),
            )

    @ssa_fn
    def test_builder_eval4(self):
        x_val = np.array([[[-1, 3, 6]], [[-1, 2, -3]], [[4, -5, 6]]], dtype=np.float32)
        with pytest.raises(ValueError, match=r"x .* rank 3"):
            v = mb.softplus_parametric(
                x=[1],
                alpha=np.array([[1, 2, 3]], dtype=np.float32),
                beta=np.array([4, 5, 6], dtype=np.float32),
            )

    @ssa_fn
    def test_builder_eval5(self):
        x_val = np.array([[[-1, 3, 6]], [[-1, 2, -3]], [[4, -5, 6]]], dtype=np.float32)
        with pytest.raises(ValueError, match=r".* dimension 1 .*"):
            v = mb.softplus_parametric(
                x=x_val,
                alpha=np.array([1, 2, 3], dtype=np.float32),
                beta=np.array([5, 6], dtype=np.float32),
            )

    @ssa_fn
    def test_builder_eval6(self):
        x_val = np.array([[[[-1, 3, 6]], [[-1, 2, -3]], [[4, -5, 6]]]], dtype=np.float32)
        with pytest.raises(ValueError, match=r"beta .* rank 1"):
            v = mb.softplus_parametric(
                x=x_val,
                alpha=np.array([1, 2, 3], dtype=np.float32),
                beta=np.array([[4, 5, 6]], dtype=np.float32),
            )

    @pytest.mark.parametrize(
        "use_cpu_only, backend, dim, chan",
        itertools.product([True, False], backends, [1, 2, 4, 8], [1, 2, 3]),
    )
    def test_builder_to_backend_stress(self, use_cpu_only, backend, dim, chan):
        shape = np.array([1, chan, dim, dim])
        x_val = np.random.rand(*shape)
        alpha_val = np.random.rand(chan).astype(np.float32)
        beta_val = np.random.rand(chan).astype(np.float32)

        input_placeholders = {"x": mb.placeholder(shape=x_val.shape)}
        input_values = {"x": x_val}

        def build(x):
            return [mb.softplus_parametric(x=x, alpha=alpha_val, beta=beta_val)]

        alpha_br = np.copy(alpha_val)
        beta_br = np.copy(beta_val)
        for i in range(1, len(x_val.shape)-1):
            alpha_br = np.expand_dims(alpha_br, i)
            beta_br = np.expand_dims(beta_br, i)
        expected_outputs = [alpha_br * np.log(np.exp(x_val * beta_br) + 1)]
        expected_output_types = [o.shape[:] + (types.fp32,) for o in expected_outputs]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs=expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
        )


class TestSoftmax:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_buidler_to_backend_smoke(self, use_cpu_only, backend):
        t = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        input_placeholders = {"x": mb.placeholder(shape=t.shape)}
        input_values = {"x": t}

        def build(x):
            return mb.softmax(x=x, axis=0)

        expected_output_types = (2, 3, types.fp32)
        expected_outputs = np.array(
            [
                [6.69285092e-03, 9.99088949e-01, 1.23394576e-04],
                [9.93307149e-01, 9.11051194e-04, 9.99876605e-01],
            ],
            dtype=np.float32,
        )
        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @ssa_fn
    def test_builder_eval(self):
        x_val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        v = mb.softmax(x=x_val, axis=0)
        assert is_close(scipy.special.softmax(x_val, axis=0), v.val)


class TestSoftsign:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        t = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        input_placeholders = {"x": mb.placeholder(shape=t.shape)}
        input_values = {"x": t}

        def build(x):
            return mb.softsign(x=x)

        expected_output_types = (2, 3, types.fp32)
        expected_outputs = np.array(
            [[-0.5, 0.66666667, -0.75], [0.8, -0.83333333, 0.85714286]],
            dtype=np.float32,
        )

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @ssa_fn
    def test_builder_eval(self):
        x_val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        v = mb.softsign(x=x_val)
        assert is_close(x_val / (1 + np.abs(x_val)), v.val)


class TestThresholdedReLU:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        t = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        input_placeholders = {"x": mb.placeholder(shape=t.shape)}
        input_values = {"x": t}

        def build(x):
            return mb.thresholded_relu(x=x, alpha=2.0)

        expected_output_types = (2, 3, types.fp32)
        expected_outputs = np.array([[0, 2, 0], [4, 0, 6]], dtype=np.float32)

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @ssa_fn
    def test_builder_eval(self):
        x_val = np.array([[0, 2, 0], [4, 0, 6]], dtype=np.float32)
        v = mb.thresholded_relu(x=x_val, alpha=2.0)
        y = x_val
        y[y < 2.0] = 0
        assert is_close(y, v.val)

    @pytest.mark.parametrize(
        "use_cpu_only, backend, dim, alpha",
        itertools.product([True], backends, [2, 4, 8], [2.0, 3.0]),
    )
    def test_builder_to_backend_stress(self, use_cpu_only, backend, dim, alpha):
        shape_x = np.array([dim, dim])
        x_val = np.random.rand(*shape_x)
        input_placeholders = {"x": mb.placeholder(shape=x_val.shape)}
        input_values = {"x": x_val}

        def build(x):
            return [mb.thresholded_relu(x=x, alpha=alpha)]

        y = x_val
        y[y < alpha] = 0
        expected_outputs = [y]
        expected_output_types = [o.shape[:] + (types.fp32,) for o in expected_outputs]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs=expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
        )



# -*- coding: utf-8 -*-

#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from __future__ import print_function as _
from __future__ import division as _
from __future__ import absolute_import as _


# A TensorArray is essentially a runtime vector<Tensor> with
#
#  - an optional requirement "infer_shape" (True by default) that all Tensors
#    stored within the vector have the same size/shape (inferred by the
#    first element stored into the tensor)
#  - an optional "element_shape" which requires all elements to have this
#    exact shape.
#  - an optional "clear_after_read" (True by default) where read of an index
#    is destructive. (It doesn't *really* destroy, but just enables a particular
#    optimization where the tensor memory can be reused).
#  - An optional "dynamic_size" (False by default) where the vector is resized
#    automatically at runtime
#
# The way it works is rather odd. To enforce "control dependency" constraints,
# a single float (flow) variable is passed between operations that write/read
# the TensorArray. Additionally, a "Resource" variable is also passed along
# which contains the actual handle to the TensorArray.
#
# The TensorArray can therefore also be passed around as as argument to while
# loops.  Thus unlike a global "Variable", this really is better thought of as
# an additional type, a list[tensor].
#
# See:
#
# https://github.com/tensorflow/tensorflow/blob/r1.6/tensorflow/python/ops/tensor_array_ops.py
# https://github.com/tensorflow/tensorflow/blob/master/tensorflow/core/kernels/tensor_array.h
# https://github.com/tensorflow/tensorflow/blob/master/tensorflow/core/kernels/tensor_array.cc
# https://github.com/tensorflow/tensorflow/blob/master/tensorflow/core/ops/data_flow_ops.cc
# https://github.com/tensorflow/tensorflow/blob/master/tensorflow/core/kernels/tensor_array_ops.cc
#
# The way we transform it is to introduce a new type. list[tensor]
# The flow variable is the list[tensor] since that is consistently passed through
# every operation.
# The 'resource' edges then gets passed as void.
#
# We would like to delete the resource edges, but once too many graph passes are
# performed, this becomes very difficult (since tuple shapes have to be updated).
# The ideal is to perform the resource edge deletion *BEFORE* any additional
# graph transformations.
# The conversion of the flow variable to list[tensor] can be performed during
# type inference.
#
#
# After this op:
# All nodes which take a TensorArray resource input will have the resource input
# edge deleted.
#
# TensorArrayV3 op will only have 1 output, a flow variable.


def tensor_array_resource_removal(gd):
    # this should be called *BEFORE* introduction of tuples,
    # and before output edges are added (for simplicity)
    for k, node in gd.items():
        if node.op.startswith("TensorArray") and node.op != "TensorArrayV3":
            # generally the resource edge is the first edge
            # input is resource, indices, flow
            # output is generally flow
            node.inputs = node.inputs[1:]

        # TensorArrayV3 node outputs resource and flow
        # shift all flow reads from TensorArray to output 0 of TensorArray
        for i in range(len(node.inputs)):
            if ":" in node.inputs[i]:
                input_node, input_index = node.inputs[i].split(":")
                input_index = int(input_index)
            else:
                input_node = node.inputs[i]
                input_index = 0
            if gd[input_node].op == "TensorArrayV3":
                if input_index == 1:
                    node.inputs[i] = "%s" % input_node



# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from . import _utils
import keras
import numpy as np


def _get_recurrent_activation_name_from_keras(activation):
    if activation == keras.activations.sigmoid:
        activation_str = "SIGMOID"
    elif activation == keras.activations.hard_sigmoid:
        activation_str = "SIGMOID_HARD"
    elif activation == keras.activations.tanh:
        activation_str = "TANH"
    elif activation == keras.activations.relu:
        activation_str = "RELU"
    elif activation == keras.activations.linear:
        activation_str = "LINEAR"
    else:
        raise NotImplementedError(
            "activation %s not supported for Recurrent layer." % activation
        )

    return activation_str


def _get_activation_name_from_keras_layer(keras_layer):
    if isinstance(keras_layer, keras.layers.advanced_activations.LeakyReLU):
        non_linearity = "LEAKYRELU"
    elif isinstance(keras_layer, keras.layers.advanced_activations.PReLU):
        non_linearity = "PRELU"
    elif isinstance(keras_layer, keras.layers.advanced_activations.ELU):
        non_linearity = "ELU"
    elif isinstance(keras_layer, keras.layers.advanced_activations.ParametricSoftplus):
        non_linearity = "PARAMETRICSOFTPLUS"
    elif isinstance(keras_layer, keras.layers.advanced_activations.ThresholdedReLU):
        non_linearity = "THRESHOLDEDRELU"
    else:
        act_name = keras_layer.activation.__name__

        if act_name == "softmax":
            non_linearity = "SOFTMAX"
        elif act_name == "sigmoid":
            non_linearity = "SIGMOID"
        elif act_name == "tanh":
            non_linearity = "TANH"
        elif act_name == "relu":
            non_linearity = "RELU"
        elif act_name == "softplus":
            non_linearity = "SOFTPLUS"
        elif act_name == "softsign":
            non_linearity = "SOFTSIGN"
        elif act_name == "hard_sigmoid":
            non_linearity = "SIGMOID_HARD"
        elif act_name == "linear":
            non_linearity = "LINEAR"
        else:
            _utils.raise_error_unsupported_categorical_option(
                "activation", act_name, "Dense", keras_layer.name  ##
            )

    return non_linearity


def _get_elementwise_name_from_keras_layer(keras_layer):
    """
    Get the keras layer name from the activation name.
    """
    mode = keras_layer.mode
    if mode == "sum":
        return "ADD"
    elif mode == "mul":
        return "MULTIPLY"
    elif mode == "concat":
        if len(keras_layer.input_shape[0]) == 3 and (
            keras_layer.concat_axis == 1 or keras_layer.concat_axis == -2
        ):
            return "SEQUENCE_CONCAT"
        elif len(keras_layer.input_shape[0]) == 4 and (
            keras_layer.concat_axis == 3 or keras_layer.concat_axis == -1
        ):
            return "CONCAT"
        elif len(keras_layer.input_shape[0]) == 2 and (
            keras_layer.concat_axis == 1 or keras_layer.concat_axis == -1
        ):
            return "CONCAT"
        else:
            option = "input_shape = %s concat_axis = %s" % (
                str(keras_layer.input_shape[0]),
                str(keras_layer.concat_axis),
            )
            _utils.raise_error_unsupported_option(option, mode, keras_layer.name)
    elif mode == "cos":
        if len(keras_layer.input_shape[0]) == 2:
            return "COS"
        else:
            option = "input_shape = %s" % (str(keras_layer.input_shape[0]))
            _utils.raise_error_unsupported_option(option, mode, keras_layer.name)
    elif mode == "dot":
        if len(keras_layer.input_shape[0]) == 2:
            return "DOT"
        else:
            option = "input_shape = %s" % (str(keras_layer.input_shape[0]))
            _utils.raise_error_unsupported_option(option, mode, keras_layer.name)
    elif mode == "max":
        return "MAX"
    elif mode == "ave":
        return "AVE"
    else:
        _utils.raise_error_unsupported_categorical_option(
            "mode", mode, "Merge", keras_layer.name
        )


def _same_elements_per_channel(x):
    """
    Test if a 3D (H,W,C) matrix x has the same element in each (H,W) matrix for each channel
    """
    eps = 1e-5
    dims = x.shape
    for c in range(dims[-1]):
        xc = x[:, :, c].flatten()
        if not np.all(np.absolute(xc - xc[0]) < eps):
            return False
    return True


def convert_dense(builder, layer, input_names, output_names, keras_layer):
    """Convert a dense layer from keras to coreml.

    Parameters
    keras_layer: layer
    ----------
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    """
    # Get input and output names
    input_name, output_name = (input_names[0], output_names[0])

    has_bias = keras_layer.bias
    # Get the weights from keras
    W = keras_layer.get_weights()[0].T
    Wb = keras_layer.get_weights()[1].T if has_bias else None

    builder.add_inner_product(
        name=layer,
        W=W,
        b=Wb,
        input_channels=keras_layer.input_dim,
        output_channels=keras_layer.output_dim,
        has_bias=has_bias,
        input_name=input_name,
        output_name=output_name,
    )


def convert_activation(builder, layer, input_names, output_names, keras_layer):
    """Convert an activation layer from keras to coreml.

    Parameters
    ----------
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    """
    # Get input and output names
    input_name, output_name = (input_names[0], output_names[0])
    non_linearity = _get_activation_name_from_keras_layer(keras_layer)

    # Add a non-linearity layer
    if non_linearity == "SOFTMAX":
        builder.add_softmax(name=layer, input_name=input_name, output_name=output_name)
        return

    params = None
    if non_linearity == "LEAKYRELU":
        params = [keras_layer.alpha]

    elif non_linearity == "PRELU":
        # In Keras 1.2  PReLU layer's weights are stored as a
        # backend tensor, not a numpy array as it claims in documentation.
        shared_axes = list(keras_layer.shared_axes)
        if not (shared_axes == [1, 2, 3] or shared_axes == [1, 2]):
            _utils.raise_error_unsupported_scenario(
                "Shared axis not being [1,2,3] " "or [1,2]", "parametric_relu", layer
            )
        params = keras.backend.eval(keras_layer.weights[0])
    elif non_linearity == "ELU":
        params = keras_layer.alpha

    elif non_linearity == "PARAMETRICSOFTPLUS":
        # In Keras 1.2  Parametric Softplus layer's weights are stored as a
        # backend tensor, not a numpy array as it claims in documentation.
        alphas = keras.backend.eval(keras_layer.weights[0])
        betas = keras.backend.eval(keras_layer.weights[1])

        if len(alphas.shape) == 3:  # (H,W,C)
            if not (
                _same_elements_per_channel(alphas) and _same_elements_per_channel(betas)
            ):
                _utils.raise_error_unsupported_scenario(
                    "Different parameter values", "parametric_softplus", layer
                )
            alphas = alphas[0, 0, :]
            betas = betas[0, 0, :]
        params = [alphas, betas]

    elif non_linearity == "THRESHOLDEDRELU":
        params = keras_layer.theta
    else:
        pass  # do nothing to parameters
    builder.add_activation(
        name=layer,
        non_linearity=non_linearity,
        input_name=input_name,
        output_name=output_name,
        params=params,
    )


def convert_merge(builder, layer, input_names, output_names, keras_layer):
    """Convert concat layer from keras to coreml.

    Parameters
    ----------
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    """
    # Get input and output names
    output_name = output_names[0]

    mode = _get_elementwise_name_from_keras_layer(keras_layer)
    builder.add_elementwise(
        name=layer, input_names=input_names, output_name=output_name, mode=mode
    )


def convert_pooling(builder, layer, input_names, output_names, keras_layer):
    """Convert pooling layer from keras to coreml.

    Parameters
    ----------
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    """
    # Get input and output names
    input_name, output_name = (input_names[0], output_names[0])

    # Pooling layer type
    if (
        isinstance(keras_layer, keras.layers.convolutional.MaxPooling2D)
        or isinstance(keras_layer, keras.layers.convolutional.MaxPooling1D)
        or isinstance(keras_layer, keras.layers.pooling.GlobalMaxPooling2D)
        or isinstance(keras_layer, keras.layers.pooling.GlobalMaxPooling1D)
    ):
        layer_type_str = "MAX"
    elif (
        isinstance(keras_layer, keras.layers.convolutional.AveragePooling2D)
        or isinstance(keras_layer, keras.layers.convolutional.AveragePooling1D)
        or isinstance(keras_layer, keras.layers.pooling.GlobalAveragePooling2D)
        or isinstance(keras_layer, keras.layers.pooling.GlobalAveragePooling1D)
    ):
        layer_type_str = "AVERAGE"
    else:
        raise TypeError("Pooling type %s not supported" % keras_layer)

    # if it's global, set the global flag
    if isinstance(keras_layer, keras.layers.pooling.GlobalMaxPooling2D) or isinstance(
        keras_layer, keras.layers.pooling.GlobalAveragePooling2D
    ):
        # 2D global pooling
        global_pooling = True
        height, width = (0, 0)
        stride_height, stride_width = (0, 0)
        padding_type = "VALID"
    elif isinstance(keras_layer, keras.layers.pooling.GlobalMaxPooling1D) or isinstance(
        keras_layer, keras.layers.pooling.GlobalAveragePooling1D
    ):
        # 1D global pooling: 1D global pooling seems problematic,
        # use this work-around
        global_pooling = False
        _, width, channels = keras_layer.input_shape
        height = 1
        stride_height, stride_width = height, width
        padding_type = "VALID"
    else:
        global_pooling = False
        # Set pool sizes and strides
        # 1D cases:
        if (
            isinstance(keras_layer, keras.layers.convolutional.MaxPooling1D)
            or isinstance(keras_layer, keras.layers.pooling.GlobalMaxPooling1D)
            or isinstance(keras_layer, keras.layers.convolutional.AveragePooling1D)
            or isinstance(keras_layer, keras.layers.pooling.GlobalAveragePooling1D)
        ):
            height, width = 1, keras_layer.pool_length
            if keras_layer.stride is not None:
                stride_height, stride_width = 1, keras_layer.stride
            else:
                stride_height, stride_width = 1, keras_layer.pool_length
        # 2D cases:
        else:
            height, width = keras_layer.pool_size
            if keras_layer.strides is None:
                stride_height, stride_width = height, width
            else:
                stride_height, stride_width = keras_layer.strides

        # Padding
        border_mode = keras_layer.border_mode
        if keras_layer.border_mode == "valid":
            padding_type = "VALID"
        elif keras_layer.border_mode == "same":
            padding_type = "SAME"
        else:
            raise TypeError("Border mode %s not supported" % border_mode)

    builder.add_pooling(
        name=layer,
        height=height,
        width=width,
        stride_height=stride_height,
        stride_width=stride_width,
        layer_type=layer_type_str,
        padding_type=padding_type,
        input_name=input_name,
        output_name=output_name,
        exclude_pad_area=True,
        is_global=global_pooling,
    )


def convert_padding(builder, layer, input_names, output_names, keras_layer):
    """Convert padding layer from keras to coreml.
    Keras only supports zero padding at this time.
    Parameters
    ----------
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    """
    # Get input and output names
    input_name, output_name = (input_names[0], output_names[0])

    if isinstance(keras_layer, keras.layers.convolutional.ZeroPadding1D):
        left, right = keras_layer.padding
        top, bottom = (0, 0)
    else:  # 2D
        top, left = keras_layer.padding
        bottom, right = keras_layer.padding

    # Now add the layer
    builder.add_padding(
        name=layer,
        left=left,
        right=right,
        top=top,
        bottom=bottom,
        value=0,
        input_name=input_name,
        output_name=output_name,
    )


def convert_cropping(builder, layer, input_names, output_names, keras_layer):
    """Convert padding layer from keras to coreml.
    Keras only supports zero padding at this time.
    Parameters
    ----------
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    """
    # Get input and output names
    input_name, output_name = (input_names[0], output_names[0])

    if isinstance(keras_layer, keras.layers.convolutional.Cropping1D):
        left, right = keras_layer.cropping
        top, bottom = (0, 0)
    else:  # 2D
        left, right = keras_layer.cropping[0]
        top, bottom = keras_layer.cropping[1]

    # Now add the layer
    builder.add_crop(
        name=layer,
        left=left,
        right=right,
        top=top,
        bottom=bottom,
        offset=[0, 0],
        input_names=[input_name],
        output_name=output_name,
    )


def convert_reshape(builder, layer, input_names, output_names, keras_layer):
    input_name, output_name = (input_names[0], output_names[0])

    input_shape = keras_layer.input_shape
    target_shape = keras_layer.target_shape

    def get_coreml_target_shape(target_shape):
        if len(target_shape) == 1:  # (D,)
            coreml_shape = (1, target_shape[0], 1, 1)
        elif len(target_shape) == 2:  # (S,D)
            coreml_shape = target_shape + (1, 1)
        elif len(target_shape) == 3:  # (H,W,C)
            coreml_shape = (1, target_shape[2], target_shape[0], target_shape[1])
        else:
            coreml_shape = None
        return coreml_shape

    def get_mode(input_shape, target_shape):
        in_shape = input_shape[1:]
        if len(in_shape) == 3 or len(target_shape) == 3:
            return 1
        else:
            return 0

    new_shape = get_coreml_target_shape(target_shape)
    if new_shape is not None:
        mode = get_mode(input_shape, target_shape)
        builder.add_reshape(
            name=layer,
            input_name=input_name,
            output_name=output_name,
            target_shape=new_shape,
            mode=mode,
        )
    else:
        _utils.raise_error_unsupported_categorical_option(
            "input_shape", str(input_shape), "reshape", layer
        )


def convert_upsample(builder, layer, input_names, output_names, keras_layer):
    """Convert convolution layer from keras to coreml.

    Parameters
    ----------
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    """
    # Get input and output names
    input_name, output_name = (input_names[0], output_names[0])

    if isinstance(keras_layer, keras.layers.convolutional.UpSampling1D):
        fh, fw = 1, keras_layer.length
    else:  # 2D
        fh, fw = keras_layer.size

    builder.add_upsample(
        name=layer,
        scaling_factor_h=fh,
        scaling_factor_w=fw,
        input_name=input_name,
        output_name=output_name,
    )


def convert_convolution(builder, layer, input_names, output_names, keras_layer):
    """Convert convolution layer from keras to coreml.

    Parameters
    ----------
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    """
    # Get input and output names
    input_name, output_name = (input_names[0], output_names[0])

    has_bias = keras_layer.bias
    is_deconv = isinstance(keras_layer, keras.layers.convolutional.Deconvolution2D)

    # Get the weights from keras.
    # Keras stores convolution weights as list of numpy arrays
    weightList = keras_layer.get_weights()
    output_shape = list(filter(None, keras_layer.output_shape))[:-1]

    # Parameter
    height, width, channels, n_filters = weightList[0].shape
    stride_height, stride_width = keras_layer.subsample

    # Weights and bias terms
    W = weightList[0]
    b = weightList[1] if has_bias else None

    # dilation factors
    dilation_factors = [1, 1]
    if isinstance(keras_layer, keras.layers.convolutional.AtrousConvolution2D):
        dilation_factors = list(keras_layer.atrous_rate)

    builder.add_convolution(
        name=layer,
        kernel_channels=channels,
        output_channels=n_filters,
        height=height,
        width=width,
        stride_height=stride_height,
        stride_width=stride_width,
        border_mode=keras_layer.border_mode,
        groups=1,
        W=W,
        b=b,
        has_bias=has_bias,
        is_deconv=is_deconv,
        output_shape=output_shape,
        input_name=input_name,
        output_name=output_name,
        dilation_factors=dilation_factors,
    )


def convert_convolution1d(builder, layer, input_names, output_names, keras_layer):
    """Convert convolution layer from keras to coreml.

    Parameters
    ----------
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    """
    # Get input and output names
    input_name, output_name = (input_names[0], output_names[0])

    has_bias = keras_layer.bias

    # Get the weights from keras.
    # Keras stores convolution weights as list of numpy arrays
    weightList = keras_layer.get_weights()
    output_shape = list(filter(None, keras_layer.output_shape))[:-1]

    # Parameter
    # weightList[0].shape = [kernel_length, input_length(time_step), input_dim, num_kernels]
    filter_length, input_length, input_dim, n_filters = weightList[0].shape
    stride_width = keras_layer.subsample[0]

    # Weights and bias terms
    W = weightList[0]
    b = weightList[1] if has_bias else None

    dilation_factors = [1, 1]
    if isinstance(keras_layer, keras.layers.convolutional.AtrousConvolution1D):
        dilation_factors[-1] = keras_layer.atrous_rate

    builder.add_convolution(
        name=layer,
        kernel_channels=input_dim,
        output_channels=n_filters,
        height=1,
        width=filter_length,
        stride_height=1,
        stride_width=stride_width,
        border_mode=keras_layer.border_mode,
        groups=1,
        W=W,
        b=b,
        has_bias=has_bias,
        is_deconv=False,
        output_shape=output_shape,
        input_name=input_name,
        output_name=output_name,
        dilation_factors=dilation_factors,
    )


def convert_lstm(builder, layer, input_names, output_names, keras_layer):
    """Convert an LSTM layer from keras to coreml.

    Parameters
    ----------
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    """

    hidden_size = keras_layer.output_dim
    input_size = keras_layer.input_shape[-1]
    if keras_layer.consume_less not in ["cpu", "gpu"]:
        raise ValueError(
            "Cannot convert Keras layer with consume_less = %s"
            % keras_layer.consume_less
        )

    output_all = keras_layer.return_sequences
    reverse_input = keras_layer.go_backwards

    # Keras: I C F O; W_x, W_h, b
    # CoreML: I F O G; W_h and W_x are separated
    W_h, W_x, b = ([], [], [])
    if keras_layer.consume_less == "cpu":
        W_h.append(keras_layer.get_weights()[1].T)
        W_h.append(keras_layer.get_weights()[7].T)
        W_h.append(keras_layer.get_weights()[10].T)
        W_h.append(keras_layer.get_weights()[4].T)

        W_x.append(keras_layer.get_weights()[0].T)
        W_x.append(keras_layer.get_weights()[6].T)
        W_x.append(keras_layer.get_weights()[9].T)
        W_x.append(keras_layer.get_weights()[3].T)

        b.append(keras_layer.get_weights()[2])
        b.append(keras_layer.get_weights()[8])
        b.append(keras_layer.get_weights()[11])
        b.append(keras_layer.get_weights()[5])
    else:
        keras_W_h = keras_layer.get_weights()[1].T
        W_h.append(keras_W_h[0 * hidden_size :][:hidden_size])
        W_h.append(keras_W_h[1 * hidden_size :][:hidden_size])
        W_h.append(keras_W_h[3 * hidden_size :][:hidden_size])
        W_h.append(keras_W_h[2 * hidden_size :][:hidden_size])

        keras_W_x = keras_layer.get_weights()[0].T
        W_x.append(keras_W_x[0 * hidden_size :][:hidden_size])
        W_x.append(keras_W_x[1 * hidden_size :][:hidden_size])
        W_x.append(keras_W_x[3 * hidden_size :][:hidden_size])
        W_x.append(keras_W_x[2 * hidden_size :][:hidden_size])

        keras_b = keras_layer.get_weights()[2]
        b.append(keras_b[0 * hidden_size :][:hidden_size])
        b.append(keras_b[1 * hidden_size :][:hidden_size])
        b.append(keras_b[3 * hidden_size :][:hidden_size])
        b.append(keras_b[2 * hidden_size :][:hidden_size])

    # Set activation type
    inner_activation_str = _get_recurrent_activation_name_from_keras(
        keras_layer.inner_activation
    )
    activation_str = _get_recurrent_activation_name_from_keras(keras_layer.activation)

    # Add to the network
    builder.add_unilstm(
        name=layer,
        W_h=W_h,
        W_x=W_x,
        b=b,
        hidden_size=hidden_size,
        input_size=input_size,
        input_names=input_names,
        output_names=output_names,
        inner_activation=inner_activation_str,
        cell_state_update_activation=activation_str,
        output_activation=activation_str,
        output_all=output_all,
        reverse_input=reverse_input,
    )


def convert_simple_rnn(builder, layer, input_names, output_names, keras_layer):
    """Convert an SimpleRNN layer from keras to coreml.

    Parameters
    ----------
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    """
    # Get input and output names
    hidden_size = keras_layer.output_dim
    input_size = keras_layer.input_shape[-1]

    output_all = keras_layer.return_sequences
    reverse_input = keras_layer.go_backwards

    if keras_layer.consume_less not in ["cpu", "gpu"]:
        raise ValueError(
            "Cannot convert Keras layer with consume_less = %s"
            % keras_layer.consume_less
        )

    W_h = np.zeros((hidden_size, hidden_size))
    W_x = np.zeros((hidden_size, input_size))
    b = np.zeros((hidden_size,))

    if keras_layer.consume_less == "cpu":
        W_h = keras_layer.get_weights()[1].T
        W_x = keras_layer.get_weights()[0].T
        b = keras_layer.get_weights()[2]
    else:
        W_h = keras_layer.get_weights()[1].T
        W_x = keras_layer.get_weights()[0].T
        b = keras_layer.get_weights()[2]

    # Set actication type
    activation_str = _get_recurrent_activation_name_from_keras(keras_layer.activation)

    # Add to the network
    builder.add_simple_rnn(
        name=layer,
        W_h=W_h,
        W_x=W_x,
        b=b,
        hidden_size=hidden_size,
        input_size=input_size,
        activation=activation_str,
        input_names=input_names,
        output_names=output_names,
        output_all=output_all,
        reverse_input=reverse_input,
    )


def convert_gru(builder, layer, input_names, output_names, keras_layer):
    """Convert a GRU layer from keras to coreml.

    Parameters
    ----------
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    """

    hidden_size = keras_layer.output_dim
    input_size = keras_layer.input_shape[-1]

    output_all = keras_layer.return_sequences
    reverse_input = keras_layer.go_backwards

    if keras_layer.consume_less not in ["cpu", "gpu"]:
        raise ValueError(
            "Cannot convert Keras layer with consume_less = %s"
            % keras_layer.consume_less
        )

    # Keras: Z R O
    # CoreML: Z R O
    W_h, W_x, b = ([], [], [])
    if keras_layer.consume_less == "cpu":
        W_x.append(keras_layer.get_weights()[0].T)
        W_x.append(keras_layer.get_weights()[3].T)
        W_x.append(keras_layer.get_weights()[6].T)

        W_h.append(keras_layer.get_weights()[1].T)
        W_h.append(keras_layer.get_weights()[4].T)
        W_h.append(keras_layer.get_weights()[7].T)

        b.append(keras_layer.get_weights()[2])
        b.append(keras_layer.get_weights()[5])
        b.append(keras_layer.get_weights()[8])
    else:
        print("consume less not implemented")

    # Set actication type
    inner_activation_str = _get_recurrent_activation_name_from_keras(
        keras_layer.inner_activation
    )
    activation_str = _get_recurrent_activation_name_from_keras(keras_layer.activation)

    # Add to the network
    builder.add_gru(
        name=layer,
        W_h=W_h,
        W_x=W_x,
        b=b,
        input_size=input_size,
        hidden_size=hidden_size,
        input_names=input_names,
        output_names=output_names,
        activation=activation_str,
        inner_activation=inner_activation_str,
        output_all=output_all,
        reverse_input=reverse_input,
    )


def convert_bidirectional(builder, layer, input_names, output_names, keras_layer):
    """Convert a bidirectional layer from keras to coreml.
        Currently assumes the units are LSTMs.

    Parameters
    ----------
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    """

    input_size = keras_layer.input_shape[-1]

    lstm_layer = keras_layer.forward_layer
    if type(lstm_layer) != keras.layers.recurrent.LSTM:
        raise TypeError("Bidirectional layers only supported with LSTM")

    if lstm_layer.go_backwards:
        raise TypeError(" 'go_backwards' mode not supported with Bidirectional layers")

    output_all = keras_layer.return_sequences

    hidden_size = lstm_layer.output_dim
    # output_size = lstm_layer.output_dim * 2

    if lstm_layer.consume_less not in ["cpu", "gpu"]:
        raise ValueError(
            "Cannot convert Keras layer with consume_less = %s"
            % keras_layer.consume_less
        )

    # Keras: I C F O; W_x, W_h, b
    # CoreML: I F O G; W_h and W_x are separated

    # Keras has all forward weights, followed by backward in the same order
    W_h, W_x, b = ([], [], [])
    if lstm_layer.consume_less == "cpu":
        W_h.append(keras_layer.get_weights()[1].T)
        W_h.append(keras_layer.get_weights()[7].T)
        W_h.append(keras_layer.get_weights()[10].T)
        W_h.append(keras_layer.get_weights()[4].T)

        W_x.append(keras_layer.get_weights()[0].T)
        W_x.append(keras_layer.get_weights()[6].T)
        W_x.append(keras_layer.get_weights()[9].T)
        W_x.append(keras_layer.get_weights()[3].T)

        b.append(keras_layer.get_weights()[2])
        b.append(keras_layer.get_weights()[8])
        b.append(keras_layer.get_weights()[11])
        b.append(keras_layer.get_weights()[5])
    else:
        keras_W_h = keras_layer.get_weights()[1].T
        W_h.append(keras_W_h[0 * hidden_size :][:hidden_size])
        W_h.append(keras_W_h[1 * hidden_size :][:hidden_size])
        W_h.append(keras_W_h[3 * hidden_size :][:hidden_size])
        W_h.append(keras_W_h[2 * hidden_size :][:hidden_size])

        keras_W_x = keras_layer.get_weights()[0].T
        W_x.append(keras_W_x[0 * hidden_size :][:hidden_size])
        W_x.append(keras_W_x[1 * hidden_size :][:hidden_size])
        W_x.append(keras_W_x[3 * hidden_size :][:hidden_size])
        W_x.append(keras_W_x[2 * hidden_size :][:hidden_size])

        keras_b = keras_layer.get_weights()[2]
        b.append(keras_b[0 * hidden_size :][:hidden_size])
        b.append(keras_b[1 * hidden_size :][:hidden_size])
        b.append(keras_b[3 * hidden_size :][:hidden_size])
        b.append(keras_b[2 * hidden_size :][:hidden_size])

    W_h_back, W_x_back, b_back = ([], [], [])
    if keras_layer.backward_layer.consume_less == "cpu":
        back_weights = keras_layer.backward_layer.get_weights()
        W_h_back.append(back_weights[1].T)
        W_h_back.append(back_weights[7].T)
        W_h_back.append(back_weights[10].T)
        W_h_back.append(back_weights[4].T)

        W_x_back.append(back_weights[0].T)
        W_x_back.append(back_weights[6].T)
        W_x_back.append(back_weights[9].T)
        W_x_back.append(back_weights[3].T)

        b_back.append(back_weights[2])
        b_back.append(back_weights[8])
        b_back.append(back_weights[11])
        b_back.append(back_weights[5])
    else:
        keras_W_h = keras_layer.backward_layer.get_weights()[1].T
        W_h_back.append(keras_W_h[0 * hidden_size :][:hidden_size])
        W_h_back.append(keras_W_h[1 * hidden_size :][:hidden_size])
        W_h_back.append(keras_W_h[3 * hidden_size :][:hidden_size])
        W_h_back.append(keras_W_h[2 * hidden_size :][:hidden_size])

        keras_W_x = keras_layer.backward_layer.get_weights()[0].T
        W_x_back.append(keras_W_x[0 * hidden_size :][:hidden_size])
        W_x_back.append(keras_W_x[1 * hidden_size :][:hidden_size])
        W_x_back.append(keras_W_x[3 * hidden_size :][:hidden_size])
        W_x_back.append(keras_W_x[2 * hidden_size :][:hidden_size])

        keras_b = keras_layer.backward_layer.get_weights()[2]
        b_back.append(keras_b[0 * hidden_size :][:hidden_size])
        b_back.append(keras_b[1 * hidden_size :][:hidden_size])
        b_back.append(keras_b[3 * hidden_size :][:hidden_size])
        b_back.append(keras_b[2 * hidden_size :][:hidden_size])

    # Set activation type
    inner_activation_str = _get_recurrent_activation_name_from_keras(
        lstm_layer.inner_activation
    )
    activation_str = _get_recurrent_activation_name_from_keras(lstm_layer.activation)

    # Add to the network
    builder.add_bidirlstm(
        name=layer,
        W_h=W_h,
        W_x=W_x,
        b=b,
        W_h_back=W_h_back,
        W_x_back=W_x_back,
        b_back=b_back,
        hidden_size=hidden_size,
        input_size=input_size,
        input_names=input_names,
        output_names=output_names,
        inner_activation=inner_activation_str,
        cell_state_update_activation=activation_str,
        output_activation=activation_str,
        output_all=output_all,
    )


def convert_batchnorm(builder, layer, input_names, output_names, keras_layer):
    """
    Parameters
    keras_layer: layer
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    """

    # Get input and output names
    input_name, output_name = (input_names[0], output_names[0])

    # Currently CoreML supports only per-channel batch-norm
    if keras_layer.mode != 0:
        raise NotImplementedError("Currently supports only per-feature normalization")

    axis = keras_layer.axis
    nb_channels = keras_layer.input_shape[axis]

    # Set parameters
    # Parameter arrangement in Keras: gamma, beta, mean, variance
    gamma = keras_layer.get_weights()[0]
    beta = keras_layer.get_weights()[1]
    mean = keras_layer.get_weights()[2]
    std = keras_layer.get_weights()[3]
    # compute adjusted parameters
    variance = std * std
    f = 1.0 / np.sqrt(std + keras_layer.epsilon)
    gamma1 = gamma * f
    beta1 = beta - gamma * mean * f
    mean[:] = 0.0  # mean
    variance[:] = 1.0 - 0.00001  # stddev

    builder.add_batchnorm(
        name=layer,
        channels=nb_channels,
        gamma=gamma1,
        beta=beta1,
        mean=mean,
        variance=variance,
        input_name=input_name,
        output_name=output_name,
    )


def convert_flatten(builder, layer, input_names, output_names, keras_layer):
    """Convert a flatten layer from keras to coreml.

    Parameters
    keras_layer: layer
    ----------
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    """
    input_name, output_name = (input_names[0], output_names[0])

    # blob_order == 0 if the input blob needs not be rearranged
    # blob_order == 1 if the input blob needs to be rearranged
    blob_order = 0

    # using keras_layer.input.shape have a "?" (Dimension[None] at the front),
    # making a 3D tensor with unknown batch size 4D
    if len(keras_layer.input.shape) == 4:
        blob_order = 1

    builder.add_flatten(
        name=layer, mode=blob_order, input_name=input_name, output_name=output_name
    )


def convert_softmax(builder, layer, input_names, output_names, keras_layer):
    """Convert a softmax layer from keras to coreml.

    Parameters
    keras_layer: layer
    ----------
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    """
    input_name, output_name = (input_names[0], output_names[0])

    builder.add_softmax(name=layer, input_name=input_name, output_name=output_name)


def convert_permute(builder, layer, input_names, output_names, keras_layer):
    """Convert a softmax layer from keras to coreml.

    Parameters
    keras_layer: layer
    ----------
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    """
    input_name, output_name = (input_names[0], output_names[0])

    keras_dims = keras_layer.dims
    # Keras permute layer index begins at 1
    if len(keras_dims) == 3:
        # Keras input tensor interpret as (H,W,C)
        x = list(np.array(keras_dims))
        i1, i2, i3 = x.index(1), x.index(2), x.index(3)
        x[i1], x[i2], x[i3] = 2, 3, 1
        # add a sequence axis
        x = [0] + x
        dim = tuple(x)
    elif len(keras_dims) == 4:
        # Here we use Keras converter as a place holder for inserting
        # permutations - the values here are not valid Keras dim parameters
        # but parameters we need to use to convert to CoreML model
        dim = keras_dims
    else:
        raise NotImplementedError("Supports only 3d permutation.")

    builder.add_permute(
        name=layer, dim=dim, input_name=input_name, output_name=output_name
    )


def convert_embedding(builder, layer, input_names, output_names, keras_layer):
    """Convert a dense layer from keras to coreml.

    Parameters
    keras_layer: layer
    ----------
        A keras layer object.

    builder: NeuralNetworkBuilder
        A neural network builder object.
    """
    # Get input and output names
    input_name, output_name = (input_names[0], output_names[0])

    # Get the weights from keras
    W = keras_layer.get_weights()[0].T

    # assuming keras embedding layers don't have biases
    builder.add_embedding(
        name=layer,
        W=W,
        b=None,
        input_dim=keras_layer.input_dim,
        output_channels=keras_layer.output_dim,
        has_bias=False,
        input_name=input_name,
        output_name=output_name,
    )


def convert_repeat_vector(builder, layer, input_names, output_names, keras_layer):
    # Keras RepeatVector only deals with 1D input
    # Get input and output names
    input_name, output_name = (input_names[0], output_names[0])

    builder.add_sequence_repeat(
        name=layer, nrep=keras_layer.n, input_name=input_name, output_name=output_name
    )


def default_skip(builder, layer, input_names, output_names, keras_layer):
    """ Layers that can be skipped (because they are train time only. """
    return



import coremltools as ct
import pytest


def _get_visible_items(d):
    return [x for x in dir(d) if not x.startswith("_")]


def _check_visible_modules(actual, expected):
    assert set(actual) == set(expected), "API mis-matched. Got %s, expected %s" % (
        actual,
        expected,
    )


class TestApiVisibilities:
    """Test public coremltools API visibilities."""

    def test_top_level(self):
        expected = [
            "ClassifierConfig",
            "EnumeratedShapes",
            "ImageType",
            "RangeDim",
            "SPECIFICATION_VERSION",
            "Shape",
            "TensorType",
            "convert",
            "converters",
            "libcoremlpython",
            "models",
            "proto",
            "target",
            "utils",
            "version",
            "test",
        ]
        if not ct.utils._is_macos():
             expected.remove("libcoremlpython")
        _check_visible_modules(_get_visible_items(ct), expected)

    def test_utils(self):
        expected = [
            "convert_double_to_float_multiarray_type",
            "convert_neural_network_spec_weights_to_fp16",
            "convert_neural_network_weights_to_fp16",
            "evaluate_classifier",
            "evaluate_classifier_with_probabilities",
            "evaluate_regressor",
            "evaluate_transformer",
            "load_spec",
            "rename_feature",
            "save_spec",
        ]
        _check_visible_modules(_get_visible_items(ct.utils), expected)

    def test_models(self):
        expected = [
            "MLModel",
            "datatypes",
            "model",
            "neural_network",
            "pipeline",
            "tree_ensemble",
            "utils",
            "nearest_neighbors",
            "feature_vectorizer",
        ]
        _check_visible_modules(_get_visible_items(ct.models), expected)

    def test_models_mlmodel(self):
        expected = [
            "author",
            "get_spec",
            "input_description",
            "license",
            "output_description",
            "predict",
            "save",
            "short_description",
            "user_defined_metadata",
            "version",
        ]
        _check_visible_modules(_get_visible_items(ct.models.MLModel), expected)

    def test_models_neural_network(self):
        expected = [
            "AdamParams",
            "NeuralNetworkBuilder",
            "SgdParams",
            "builder",
            "datatypes",
            "flexible_shape_utils",
            "optimization_utils",
            "printer",
            "quantization_utils",
            "set_training_features",
            "set_transform_interface_params",
            "spec_inspection_utils",
            "update_optimizer_utils",
            "utils",
        ]
        _check_visible_modules(_get_visible_items(ct.models.neural_network), expected)

    def test_models_neural_network_utils(self):
        expected = ["NeuralNetworkBuilder", "make_image_input", "make_nn_classifier"]
        _check_visible_modules(
            _get_visible_items(ct.models.neural_network.utils), expected
        )

    def test_models_tree_ensemble(self):
        expected = [
            "TreeEnsembleBase",
            "TreeEnsembleClassifier",
            "TreeEnsembleRegressor",
            "set_classifier_interface_params",
            "set_regressor_interface_params",
        ]
        _check_visible_modules(_get_visible_items(ct.models.tree_ensemble), expected)

    def test_models_pipeline(self):
        expected = [
            "Pipeline",
            "PipelineClassifier",
            "PipelineRegressor",
            "set_classifier_interface_params",
            "set_regressor_interface_params",
            "set_training_features",
            "set_transform_interface_params",
        ]
        _check_visible_modules(_get_visible_items(ct.models.pipeline), expected)

    def test_converters(self):
        expected = [
            "ClassifierConfig",
            "EnumeratedShapes",
            "ImageType",
            "RangeDim",
            "Shape",
            "TensorType",
            "caffe",
            "convert",
            "keras",
            "libsvm",
            "mil",
            "onnx",
            "sklearn",
            "xgboost",
        ]
        _check_visible_modules(_get_visible_items(ct.converters), expected)

    def test_converters_caffe(self):
        _check_visible_modules(_get_visible_items(ct.converters.caffe), ["convert"])

    @pytest.mark.skipif(
        ct.utils._python_version() >= (3, 8, 0),
        reason="Keras isn't compatible with Python 3.8+.",
    )
    @pytest.mark.xfail(
         condition=not ct.utils._is_macos(),
         reason="rdar://65138103 (Keras converter not exposed on Linux)",
         run=False,
     )
    def test_converters_keras(self):
        _check_visible_modules(_get_visible_items(ct.converters.keras), ["convert"])

    def test_converters_libsvm(self):
        _check_visible_modules(_get_visible_items(ct.converters.libsvm), ["convert"])

    @pytest.mark.skipif(
        ct.utils._python_version() >= (3, 8, 0),
        reason="ONNX isn't compatible with Python 3.8+.",
    )
    def test_converters_onnx(self):
        _check_visible_modules(_get_visible_items(ct.converters.onnx), ["convert"])

    def test_converters_sklearn(self):
        _check_visible_modules(_get_visible_items(ct.converters.sklearn), ["convert"])

    def test_converters_xgboost(self):
        _check_visible_modules(_get_visible_items(ct.converters.xgboost), ["convert"])

    def test_converters_mil(self):
        pass  # TODO: [Create API visibility tests for MIL](rdar://64413959)

    def test_models_neural_network_quantization_utils(self):
        expected = [
            "AdvancedQuantizedLayerSelector",
            "MatrixMultiplyLayerSelector",
            "ModelMetrics",
            "NoiseMetrics",
            "OutputMetric",
            "QuantizedLayerSelector",
            "TopKMetrics",
            "activate_int8_int8_matrix_multiplications",
            "compare_models",
            "quantize_weights",
        ]
        _check_visible_modules(
            _get_visible_items(ct.models.neural_network.quantization_utils), expected
        )

    def test_models_neural_network_flexible_shape_utils(self):
        expected = [
            "NeuralNetworkImageSize",
            "NeuralNetworkImageSizeRange",
            "NeuralNetworkMultiArrayShape",
            "NeuralNetworkMultiArrayShapeRange",
            "Shape",
            "ShapeRange",
            "Size",
            "add_enumerated_image_sizes",
            "add_enumerated_multiarray_shapes",
            "add_multiarray_ndshape_enumeration",
            "set_multiarray_ndshape_range",
            "update_image_size_range",
            "update_multiarray_shape_range",
        ]
        _check_visible_modules(
            _get_visible_items(ct.models.neural_network.flexible_shape_utils), expected
        )

    def test_models_neural_network_update_optimizer_utils(self):
        expected = ["AdamParams", "Batch", "RangeParam", "SgdParams"]
        _check_visible_modules(
            _get_visible_items(ct.models.neural_network.update_optimizer_utils),
            expected,
        )

    def test_models_neural_network_optimization_utils(self):
        _check_visible_modules(
            _get_visible_items(ct.models.neural_network.optimization_utils), [],
        )



from __future__ import absolute_import as _
from __future__ import division as _
from __future__ import print_function as _

import unittest

from coremltools._deps import _HAS_ONNX, MSG_ONNX_NOT_FOUND

if _HAS_ONNX:
    import onnx
    from ._test_utils import _onnx_create_model
    from onnx import helper, ModelProto, TensorProto
    from coremltools.converters.onnx import convert
from coremltools.proto import NeuralNetwork_pb2  # type: ignore


def _make_model_acos_exp_topk():  # type: (...) -> ModelProto
    """
  make a very simple model for testing: input->clip->exp->topk->2 outputs
  """
    inputs = [("input0", (10,), TensorProto.FLOAT), ("K", (1,), TensorProto.INT64)]
    outputs = [
        ("output_values", (3,), TensorProto.FLOAT),
        ("output_indices", (3,), TensorProto.INT64),
    ]
    acos = helper.make_node("Acos", inputs=[inputs[0][0]], outputs=["acos_out"])
    exp = helper.make_node("Exp", inputs=[acos.output[0]], outputs=["exp_out"])
    topk = helper.make_node(
        "TopK",
        inputs=[exp.output[0], inputs[1][0]],
        outputs=[outputs[0][0], outputs[1][0]],
        axis=0,
    )
    return _onnx_create_model([acos, exp, topk], inputs, outputs)


def _make_model_flatten_axis3():  # type: (...) -> ModelProto
    """
  make a simple model: 4-D input -> flatten (axis=3)-> output
  """
    inputs = [("input", (1, 3, 10, 20), TensorProto.FLOAT)]
    outputs = [("output", (30, 20), TensorProto.FLOAT)]
    flatten = helper.make_node(
        "Flatten", inputs=[inputs[0][0]], outputs=[outputs[0][0]], axis=3
    )
    return _onnx_create_model([flatten], inputs, outputs)


@unittest.skipUnless(_HAS_ONNX, MSG_ONNX_NOT_FOUND)
class CustomLayerTest(unittest.TestCase):
    def test_unsupported_ops(self):  # type: () -> None

        onnx_model = _make_model_acos_exp_topk()
        coreml_model = convert(onnx_model, add_custom_layers=True)

        spec = coreml_model.get_spec()
        layers = spec.neuralNetwork.layers
        self.assertIsNotNone(layers[0].custom)
        self.assertIsNotNone(layers[2].custom)
        self.assertEqual("Acos", layers[0].custom.className)
        self.assertEqual("TopK", layers[2].custom.className)

    def test_unsupported_ops_provide_functions(self):  # type: () -> None
        def convert_acos(builder, node, graph, err):
            params = NeuralNetwork_pb2.CustomLayerParams()
            params.className = node.op_type
            params.description = "Custom layer that corresponds to the ONNX op {}".format(
                node.op_type,
            )

            builder.add_custom(
                name=node.name,
                input_names=node.inputs,
                output_names=node.outputs,
                custom_proto_spec=params,
            )

        def convert_topk(builder, node, graph, err):
            params = NeuralNetwork_pb2.CustomLayerParams()
            params.className = node.op_type
            params.description = "Custom layer that corresponds to the ONNX op {}".format(
                node.op_type,
            )
            params.parameters["axis"].intValue = node.attrs.get("axis", -1)

            builder.add_custom(
                name=node.name,
                input_names=node.inputs,
                output_names=node.outputs,
                custom_proto_spec=params,
            )

        onnx_model = _make_model_acos_exp_topk()
        coreml_model = convert(
            model=onnx_model,
            add_custom_layers=True,
            custom_conversion_functions={"Acos": convert_acos, "TopK": convert_topk},
        )

        spec = coreml_model.get_spec()
        layers = spec.neuralNetwork.layers
        self.assertIsNotNone(layers[0].custom)
        self.assertIsNotNone(layers[2].custom)
        self.assertEqual("Acos", layers[0].custom.className)
        self.assertEqual("TopK", layers[2].custom.className)
        self.assertEqual(0, layers[2].custom.parameters["axis"].intValue)

    def test_node_name_type_custom_functions(self):  # type: () -> None
        def convert_acos(builder, node, graph, err):
            params = NeuralNetwork_pb2.CustomLayerParams()
            params.className = node.op_type
            params.description = "Custom layer that corresponds to the ONNX op {}".format(
                node.op_type,
            )

            builder.add_custom(
                name=node.name,
                input_names=node.inputs,
                output_names=node.outputs,
                custom_proto_spec=params,
            )

        def convert_topk_generic(builder, node, graph, err):
            params = NeuralNetwork_pb2.CustomLayerParams()
            params.className = node.op_type
            params.description = "Custom layer that corresponds to the ONNX op {}".format(
                node.op_type,
            )
            params.parameters["axis"].intValue = node.attrs.get("axis", -1)
            params.parameters["k"].intValue = node.attrs["k"]

            builder.add_custom(
                name=node.name,
                input_names=node.inputs,
                output_names=node.outputs,
                custom_proto_spec=params,
            )

        def convert_topk_node_specific(builder, node, graph, err):
            params = NeuralNetwork_pb2.CustomLayerParams()
            params.className = node.op_type
            params.description = "Custom layer that corresponds to the ONNX op {}".format(
                node.op_type,
            )
            params.parameters["axis"].intValue = node.attrs.get("axis", -1)

            builder.add_custom(
                name=node.name,
                input_names=node.inputs,
                output_names=node.outputs,
                custom_proto_spec=params,
            )

        onnx_model = _make_model_acos_exp_topk()
        coreml_model = convert(
            model=onnx_model,
            add_custom_layers=True,
            custom_conversion_functions={
                "Acos": convert_acos,
                "TopK": convert_topk_generic,
                "output_values_output_indices": convert_topk_node_specific,
            },
        )

        spec = coreml_model.get_spec()
        layers = spec.neuralNetwork.layers
        self.assertIsNotNone(layers[0].custom)
        self.assertIsNotNone(layers[2].custom)
        self.assertEqual("Acos", layers[0].custom.className)
        self.assertEqual("TopK", layers[2].custom.className)
        self.assertEqual(0, layers[2].custom.parameters["axis"].intValue)

    def test_unsupported_op_attribute(self):  # type: () -> None
        onnx_model = _make_model_flatten_axis3()
        coreml_model = convert(onnx_model, add_custom_layers=True)

        spec = coreml_model.get_spec()
        layers = spec.neuralNetwork.layers
        self.assertIsNotNone(layers[0].custom)
        self.assertEqual("Flatten", layers[0].custom.className)

    def test_unsupported_op_attribute_provide_functions(self):  # type: () -> None
        def convert_flatten(builder, node, graph, err):
            params = NeuralNetwork_pb2.CustomLayerParams()
            params.className = node.op_type
            params.description = "Custom layer that corresponds to the ONNX op {}".format(
                node.op_type,
            )
            params.parameters["axis"].intValue = node.attrs["axis"]

            builder.add_custom(
                name=node.name,
                input_names=node.inputs,
                output_names=node.outputs,
                custom_proto_spec=params,
            )

        def test_conversion(onnx_model, add_custom_layers=False):
            coreml_model = convert(
                onnx_model,
                add_custom_layers=add_custom_layers,
                custom_conversion_functions={"Flatten": convert_flatten},
            )

            spec = coreml_model.get_spec()
            layers = spec.neuralNetwork.layers
            self.assertIsNotNone(layers[0].custom)
            self.assertEqual("Flatten", layers[0].custom.className)
            self.assertEqual(3, layers[0].custom.parameters["axis"].intValue)

        onnx_model = _make_model_flatten_axis3()
        # Test with add_custom_layers True
        convert(
            onnx_model,
            add_custom_layers=True,
            custom_conversion_functions={"Flatten": convert_flatten},
        )

        # Test with add_custom_layers False
        convert(
            onnx_model,
            add_custom_layers=False,
            custom_conversion_functions={"Flatten": convert_flatten},
        )



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from coremltools.converters.mil import testing_reqs
from coremltools.converters.mil.testing_reqs import *
from coremltools.converters.mil.frontend.tensorflow.test.testing_utils import (
    make_tf_graph,
    TensorFlowBaseTest
)

# Custom Op imports
from coremltools.converters.mil.frontend.tensorflow.tf_op_registry import register_tf_op

# Importing _TF_OPS_REGISTRY to ensure `overriding` existing TF op does not break
# testing of default op
# pytest imports all the tests and hence overriding op invokes custom op which is not expected
# In real usecase, importing following is not recommended!!
from coremltools.converters.mil.frontend.tensorflow.tf_op_registry import (
    _TF_OPS_REGISTRY,
)
from coremltools.converters.mil.mil.ops.defs._op_reqs import *
from coremltools.converters.mil.mil import Builder as mb


class TestCompositeOp(TensorFlowBaseTest):
    @pytest.fixture(scope="class")
    def create_custom_selu(self):
        default_selu = _TF_OPS_REGISTRY.get("Selu", None)

        @register_tf_op(tf_alias=[], override=True)
        def Selu(context, node):
            x = context[node.inputs[0]]
            alpha = 1.6732631921768188
            lamda = 1.0507010221481323
            out_elu = mb.elu(x=x, alpha=alpha)
            out = mb.mul(x=out_elu, y=lamda, name=node.name)
            context.add(node.name, out)

        yield

        _TF_OPS_REGISTRY["Selu"] = default_selu

    @pytest.mark.parametrize(
        "use_cpu_only, backend, rank",
        itertools.product([True, False], backends, list(range(1, 5))),
    )
    @pytest.mark.usefixtures("create_custom_selu")
    def test_selu(self, use_cpu_only, backend, rank):
        input_shape = np.random.randint(low=1, high=6, size=rank)

        @make_tf_graph([input_shape])
        def build_model(x):
            return tf.keras.activations.selu(x)

        model, inputs, outputs = build_model

        input_values = [random_gen(input_shape, -10.0, 10.0)]
        input_dict = dict(zip(inputs, input_values))
        TensorFlowBaseTest.run_compare_tf(
            model,
            input_dict,
            outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )
